# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.

from __future__ import absolute_import
import time
import calendar

from .protocol import (
    to_hstring,
    read_first_non_ws,
    read_int_and_next,
    read_int,
    read_float)

from .errors import (
    ProtocolError)
from six.moves import filter
import six
from six.moves import range


# Constants for Misc-Info (needed in requests below)

MI_RECPT=0
MI_CC_RECPT=1
MI_COMM_TO=2
MI_COMM_IN=3
MI_FOOTN_TO=4
MI_FOOTN_IN=5
MI_LOC_NO=6
MI_REC_TIME=7
MI_SENT_BY=8
MI_SENT_AT=9
MI_BCC_RECPT=15

MIR_TO = MI_RECPT
MIR_CC = MI_CC_RECPT
MIR_BCC = MI_BCC_RECPT

MIC_COMMENT = MI_COMM_TO
MIC_FOOTNOTE = MI_FOOTN_TO


# TODO: Rename the to_string() to something better. Its purpose is to
# return the serialized data.

class EmptyResponse(object):
    @classmethod
    def parse(cls, buf):
        return None

class String(bytes):
    def __new__(cls, s=None):
        """
        @param s Encoded string (not unicode)
        """
        if isinstance(s, six.text_type):
            # For convenience. Latin-1 is the default-encoding in
            # LysKOM
            try:
                s = s.encode('latin-1')
            except UnicodeEncodeError:
                raise ValueError("Failed to encode {!r} to latin-1".format(s))

        if s is None:
            return bytes.__new__(cls)
        else:
            return bytes.__new__(cls, s)

    @classmethod
    def parse(cls, buf):
        # Parse a string (Hollerith notation)
        (length, h) = read_int_and_next(buf)
        if h != b"H":
            raise ProtocolError()
        return cls(buf.receive_string(length))

    def to_string(self):
        if isinstance(self, six.text_type):
            raise ValueError("Un-encoded string: {!r}".format(self))
        return b"%dH%s" % (len(self), self)

class Float(float):
    @classmethod
    def parse(cls, buf):
        return read_float(buf)

    def to_string(self):
        raise NotImplementedError()

class Int(int):
    @classmethod
    def parse(cls, buf):
        return read_int(buf)

    def to_string(self):
        return b"%d" % (self,)

class Bool(Int):
    # todo: inherit from bool instead of int
    pass

class Int8(Int):
    pass

class Int16(Int):
    pass

class Int32(Int):
    pass

class AuxNo(Int32):
    pass

class ConfNo(Int16):
    pass

class PersNo(ConfNo):
    pass

class TextNo(Int32):
    pass

class LocalTextNo(Int32):
    pass

class SessionNo(Int32):
    pass

class GarbNice(Int32):
    pass

class InfoType(Int32):
    # TODO: Should only allow values of MiscInfo.
    # http://www.lysator.liu.se/lyskom/protocol/11.1/protocol-a.html#Info-Type
    pass


class Array(list):
    """Sub-class this to use it.
    """
    ELEMENT_CLASS = None # Must be set in subclass

    def __init__(self, iterable=None):
        if self.ELEMENT_CLASS is None:
            raise ValueError("No element class specified")
        if iterable is None:
            list.__init__(self)
        else:
            iterable = [ self.ELEMENT_CLASS(v) for v in iterable ]
            list.__init__(self, iterable)

    def __setitem__(self, i, y):
        if not isinstance(y, self.ELEMENT_CLASS):
            y = self.ELEMENT_CLASS(y)
        return list.__setitem__(self, i, y)

    def append(self, x):
        if not isinstance(x, self.ELEMENT_CLASS):
            x = self.ELEMENT_CLASS(x)
        return list.append(self, x)

    def insert(self, i, x):
        if not isinstance(x, self.ELEMENT_CLASS):
            x = self.ELEMENT_CLASS(x)
        return list.insert(self, i, x)

    def extend(self, l):
        return list.extend(self, self.__class__(l))

    def __add__(self, other):
        return self.__class__(list.__add__(self, other))

    #def __repr__(self):
    #    return "{:s}({:s})".format(
    #        self.__class__.__name__,
    #        list.__repr__(self))

    @classmethod
    def parse(cls, buf):
        length = read_int(buf)
        obj = cls()
        left = read_first_non_ws(buf)
        if left == b"*":
            # Empty or special case of unwanted data
            return obj
        elif left != b"{":
            raise ProtocolError()
        for i in range(0, length):
            el = cls.ELEMENT_CLASS.parse(buf)
            obj.append(el)
        right = read_first_non_ws(buf)
        if right != b"}":
            raise ProtocolError()
        return obj

    def to_string(self):
        self._validate_array()
        if len(self) > 0:
            return b"%d { %s }" % (len(self), b" ".join([x.to_string() for x in self]))
        else:
            return b"0 { }"

    def _validate_array(self):
        for v in self:
            if not isinstance(v, self.ELEMENT_CLASS):
                raise ValueError("Array of {!r} contains invalid element ({!r})".format(
                        self.ELEMENT_CLASS, v))


class ArrayInt32(Array):
    ELEMENT_CLASS = Int32

class ArrayLocalTextNo(Array):
    ELEMENT_CLASS = LocalTextNo

class ArrayTextNo(Array):
    ELEMENT_CLASS = TextNo

class ArrayString(Array):
    ELEMENT_CLASS = String

class Bitstring(list):
    """Some type of base class. Not meant to be used directly as datatype.
    """
    LENGTH = None # Must be set in subclass

    def __init__(self, iterable=None):
        length = self.LENGTH
        if length is None:
            raise ValueError("No length specified")
        if length < 1:
            raise ValueError("Cannot be empty")
        if iterable is None:
            iterable = [0]*length
        if len(iterable) != length:
            raise ValueError("Wrong length, expected {:d}".format(length))
        list.__init__(self, iterable)
        self._validate_bitstring()

    @classmethod
    def parse(cls, buf):
        obj = cls()
        length = cls.LENGTH
        char = read_first_non_ws(buf)
        for i in range(0, length):
            if char == b"0":
                obj[i] = 0
            elif char == b"1":
                obj[i] = 1
            else:
                raise ProtocolError()
            char = buf.receive_char()
        return obj

    def to_string(self):
        self._validate_bitstring()
        return (b"%d"*self.LENGTH) % tuple(self)

    def _validate_bitstring(self):
        assert len(self) == self.LENGTH
        for v in self:
            if v not in (0, 1):
                raise ValueError("Bitstring values can only be 0 or 1 (got {!r})".format(v))
        

class Bitstring4(Bitstring):
    LENGTH = 4

class Bitstring8(Bitstring):
    LENGTH = 8

class Bitstring16(Bitstring):
    LENGTH = 16

def _create_bitstring_accessors(index):
    """For creating named properties for certain list indicies.

    Usage: original = property(*_create_bitstring_accessors(0))
    
    """
    assert index >= 0
    def _create_get(index):
        def get_wrapper(self):
            return self.__getitem__(index)
        return get_wrapper
    def _create_set(index):
        def set_wrapper(self, value):
            return self.__setitem__(index, value)
        return set_wrapper

    return (_create_get(index), _create_set(index))



# TIME

class Time(object):
    """Assumes all dates are in UTC timezone.
    """
    def __init__(self, seconds=0, minutes=0, hours=0, day=0, month=0, year=0,
                 day_of_week=0, day_of_year=0, is_dst=0, ptime=None):
        if ptime is None:
            self.seconds = seconds
            self.minutes = minutes
            self.hours = hours
            self.day = day
            self.month = month # 0 .. 11 
            self.year = year # no of years since 1900
            self.day_of_week = day_of_week # 0 = Sunday ... 6 = Saturday
            self.day_of_year = day_of_year # 0 ... 365
            self.is_dst = is_dst
        else:
            (dy,dm,dd,th,tm,ts, wd, yd, dt) = time.gmtime(ptime)
            self.seconds = ts
            self.minutes = tm
            self.hours = th
            self.day = dd
            self.month = dm -1 
            self.year = dy - 1900 
            self.day_of_week = (wd + 1) % 7
            self.day_of_year = yd - 1
            self.is_dst = dt

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.seconds = Int32.parse(buf)
        obj.minutes = Int32.parse(buf)
        obj.hours = Int32.parse(buf)
        obj.day = Int32.parse(buf)
        obj.month = Int32.parse(buf)
        obj.year = Int32.parse(buf)
        obj.day_of_week = Int32.parse(buf)
        obj.day_of_year = Int32.parse(buf)
        obj.is_dst = Bool.parse(buf)
        return obj

    def __repr__(self):
        return ("Time(seconds={sec!r}, minutes={min!r}, hours={hours!r}, day={day!r}, "
                "month={month!r}, year={year!r}, day_of_week={dow!r}, day_of_year={doy!r}, "
                "is_dst={dst!r})".format(
                sec=self.seconds,
                min=self.minutes,
                hours=self.hours,
                day=self.day,
                month=self.month,
                year=self.year,
                dow=self.day_of_week,
                doy=self.day_of_year,
                dst=self.is_dst))

    def to_string(self):
        return b"%d %d %d %d %d %d %d %d %d" % (
            self.seconds,
            self.minutes,
            self.hours,
            self.day,
            self.month,
            self.year,
            self.day_of_week, # ignored by server
            self.day_of_year, # ignored by server
            self.is_dst)

    def to_python_time(self):
        return calendar.timegm((self.year + 1900,
                                self.month + 1,
                                self.day,
                                self.hours,
                                self.minutes,
                                self.seconds,
                                (self.day_of_week - 1) % 7,
                                self.day_of_year + 1,
                                self.is_dst))

    def to_date_and_time(self):
        return "%04d-%02d-%02d %02d:%02d:%02d" % \
            (self.year + 1900, self.month + 1, self.day,
             self.hours, self.minutes, self.seconds)

    def to_iso_8601(self):
        """Example: 1994-11-05T13:15:30Z"""
        return "%04d-%02d-%02dT%02d:%02d:%02dZ" % (
            self.year + 1900, self.month + 1, self.day,
            self.hours, self.minutes, self.seconds)

    def __str__(self):
        return "<Time %s, dst=%d>" % (self.to_date_and_time(), self.is_dst)

    def __eq__(self, other):
        return (self.seconds == other.seconds and
                self.minutes == other.minutes and
                self.hours == other.hours and
                self.day == other.day and
                self.month == other.month and
                self.year == other.year and
                self.day_of_week == other.day_of_week and
                self.day_of_year == other.day_of_year and
                self.is_dst == other.is_dst)

    def __ne__(self, other):
        return not self == other


# RESULT FROM LOOKUP-Z-NAME

class ConfZInfo(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.name = String.parse(buf)
        obj.type = ConfType.parse(buf)
        obj.conf_no = ConfNo.parse(buf)
        return obj

    def __str__(self):
        return "<ConfZInfo %d: %s>" % \
            (self.conf_no, self.name)

    def __eq__(self, other):
        return (self.name == other.name and
                self.type == other.type and
                self.conf_no == other.conf_no)

    def __ne__(self, other):
        return not self == other

# RAW MISC-INFO (AS IT IS IN PROTOCOL A)

class RawMiscInfo(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.type = Int32.parse(buf)
        if obj.type in [MI_REC_TIME, MI_SENT_AT]:
            obj.data = Time.parse(buf)
        else:
            obj.data = Int32.parse(buf)
        return obj

    def __str__(self):
        return "<MiscInfo %d: %s>" % (self.type, self.data)

    def __eq__(self, other):
        return (self.type == other.type and
                self.data == other.data)

    def __ne__(self, other):
        return not self == other

class ArrayRawMiscInfo(Array):
    ELEMENT_CLASS = RawMiscInfo


class MIRecipient(object):
    def __init__(self, type=MIR_TO, recpt=0):
        self.type = type # MIR_TO, MIR_CC or MIR_BCC
        self.recpt = recpt   # Always present
        self.loc_no = None   # Always present
        self.rec_time = None # Will be None if not sent by server
        self.sent_by = None  # Will be None if not sent by server
        self.sent_at = None  # Will be None if not sent by server

    def decode_additional(self, raw, i):
        while i < len(raw):
            if raw[i].type == MI_LOC_NO:
                self.loc_no = raw[i].data
            elif raw[i].type == MI_REC_TIME:
                self.rec_time = raw[i].data
            elif raw[i].type == MI_SENT_BY:
                self.sent_by = raw[i].data
            elif raw[i].type == MI_SENT_AT:
                self.sent_at = raw[i].data
            else:
                return i 
            i = i + 1
        return i

    def get_tuples(self):
        return [(self.type, self.recpt)]

    def __eq__(self, other):
        return (self.type == other.type and
                self.recpt == other.recpt and
                self.loc_no == other.loc_no and
                self.rec_time == other.rec_time and
                self.sent_by == other.send_by and
                self.sent_at == other.send_at)

    def __ne__(self, other):
        return not self == other

class MICommentTo(object):
    def __init__(self, type = MIC_COMMENT, text_no = 0):
        self.type = type
        self.text_no = text_no
        self.sent_by = None
        self.sent_at = None
        
    def decode_additional(self, raw, i):
        while i < len(raw):
            if raw[i].type == MI_SENT_BY:
                self.sent_by = raw[i].data
            elif raw[i].type == MI_SENT_AT:
                self.sent_at = raw[i].data
            else:
                return i 
            i = i + 1
        return i

    def get_tuples(self):
        return [(self.type, self.text_no)]

    def __eq__(self, other):
        return (self.type == other.type and
                self.text_no == other.text_no and
                self.sent_by == other.send_by and
                self.sent_at == other.send_at)

    def __ne__(self, other):
        return not self == other

class MICommentIn(object):
    def __init__(self, type = MIC_COMMENT, text_no = 0):
        self.type = type
        self.text_no = text_no

    def get_tuples(self):
        # Cannot send these to sever
        return []

    def __eq__(self, other):
        return (self.type == other.type and
                self.text_no == other.text_no)

    def __ne__(self, other):
        return not self == other

# COOKED MISC-INFO (MORE TASTY)
# N.B: This class represents the whole array, not just one item
class CookedMiscInfo(object):
    def __init__(self, other=None):
        if other is None:
            self.recipient_list = []
            self.comment_to_list = []
            self.comment_in_list = []
        else:
            self.recipient_list = other.recipient_list
            self.comment_to_list = other.comment_to_list
            self.comment_in_list = other.comment_in_list

    @classmethod
    def parse(cls, buf):
        obj = cls()
        raw = ArrayRawMiscInfo.parse(buf)
        i = 0
        while i < len(raw):
            if raw[i].type in [MI_RECPT, MI_CC_RECPT, MI_BCC_RECPT]:
                r = MIRecipient(raw[i].type, raw[i].data)
                i = r.decode_additional(raw, i+1)
                obj.recipient_list.append(r)
            elif raw[i].type in [MI_COMM_TO, MI_FOOTN_TO]:
                ct = MICommentTo(raw[i].type, raw[i].data)
                i = ct.decode_additional(raw, i+1)
                obj.comment_to_list.append(ct)
            elif raw[i].type in [MI_COMM_IN, MI_FOOTN_IN]:
                ci = MICommentIn(raw[i].type - 1 , raw[i].data  ) # KLUDGE :-)
                i = i + 1
                obj.comment_in_list.append(ci)
            else:
                raise ProtocolError
        return obj

    def to_string(self):
        l = []
        for r in self.comment_to_list + \
            self.recipient_list + \
            self.comment_in_list:
            l = l + r.get_tuples()
        return b"%d { %s}" % (len(l),
                              b"".join([b"%d %d " % \
                                            (x[0], x[1]) for x in l]))


    def __eq__(self, other):
        return (self.recipient_list == other.recipient_list and
                self.comment_to_list == other.comment_to_list and
                self.comment_in_list == other.comment_in_list)

    def __ne__(self, other):
        return not self == other

# AUX INFO

class AuxItemFlags(Bitstring8):
    deleted = property(*_create_bitstring_accessors(0))
    inherit = property(*_create_bitstring_accessors(1))
    secret = property(*_create_bitstring_accessors(2))
    hide_creator = property(*_create_bitstring_accessors(3))
    dont_garb = property(*_create_bitstring_accessors(4))
    reserved2 = property(*_create_bitstring_accessors(5))
    reserved3 = property(*_create_bitstring_accessors(6))
    reserved4 = property(*_create_bitstring_accessors(7))
        

class AuxItem(object): 
    def __init__(self, aux_item=None):
        if aux_item is not None:
            self.aux_no = aux_item.aux_no
            self.tag = aux_item.tag
            self.creator = aux_item.creator
            self.created_at = aux_item.created_at
            self.flags = aux_item.flags
            self.inherit_limit = aux_item.inherit_limit
            self.data = aux_item.data
        else:
            self.aux_no = 0
            self.tag = 0
            self.creator = 0
            self.created_at = Time()
            self.flags = AuxItemFlags()
            self.inherit_limit = 0
            self.data = ""

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.aux_no = AuxNo.parse(buf)
        obj.tag = Int32.parse(buf)
        obj.creator = PersNo.parse(buf)
        obj.created_at = Time.parse(buf)
        obj.flags = AuxItemFlags.parse(buf)
        obj.inherit_limit = Int32.parse(buf)
        obj.data = String.parse(buf)
        return obj

    def __str__(self):
        return "<AuxItem %d>" % self.tag

    def __eq__(self, other):
        return (self.aux_no == other.aux_no and
                self.tag == other.tag and
                self.creator == other.creator and
                self.created_at == other.created_at and
                self.flags == other.flags and
                self.inherit_limit == other.inherit_limit and
                self.data == other.data)

    def __ne__(self, other):
        return not self == other


class AuxItemInput(object): 
    def __init__(self, aux_item=None, tag=0, flags=None, inherit_limit=0, data=None):
        if aux_item is not None:
            self.tag = aux_item.tag
            self.flags = aux_item.flags
            self.inherit_limit = aux_item.inherit_limit
            self.data = aux_item.data
        else:
            if flags is None:
                flags = AuxItemFlags()
            if data is None:
                data = ""
            self.tag = tag
            self.flags = flags
            self.inherit_limit = inherit_limit
            self.data = data

    def __str__(self):
        return "<AuxItemInput %d>" % self.tag

    def to_string(self):
        return b"%d %s %d %s" % \
            (self.tag,
             self.flags.to_string(),
             self.inherit_limit,
             to_hstring(self.data))

    def __eq__(self, other):
        return (self.tag == other.tag and
                self.flags == other.flags and
                self.inherit_limit == other.inherit_limit and
                self.data == other.data)

    def __ne__(self, other):
        return not self == other

class ArrayAuxItem(Array):
    ELEMENT_CLASS = AuxItem

class ArrayAuxItemInput(Array):
    ELEMENT_CLASS = AuxItemInput


# Functions operating on lists of AuxItems

def all_aux_items_with_tag(ail, tag):
    return list(filter(lambda x, tag=tag: x.tag == tag, ail))
     
def first_aux_items_with_tag(ail, tag):
    all = all_aux_items_with_tag(ail, tag)
    if len(all) == 0:
        return None
    else:
        return all[0]
     
# TEXT

class TextStat(object):
    def __init__(self, creation_time=None, author=0, no_of_lines=0, no_of_chars=0,
                 no_of_marks=0, misc_info=None, aux_items=None):
        self.creation_time = creation_time
        self.author = author
        self.no_of_lines = no_of_lines
        self.no_of_chars = no_of_chars
        self.no_of_marks = no_of_marks
        if misc_info is None:
            misc_info = CookedMiscInfo()
        self.misc_info = misc_info
        if aux_items is None:
            aux_items = []
        self.aux_items = aux_items

    @classmethod
    def parse(cls, buf, old_format=0):
        obj = cls()
        obj.creation_time = Time.parse(buf)
        obj.author = PersNo.parse(buf)
        obj.no_of_lines = Int32.parse(buf)
        obj.no_of_chars = Int32.parse(buf)
        obj.no_of_marks = Int16.parse(buf)
        obj.misc_info = CookedMiscInfo.parse(buf)
        if old_format:
            obj.aux_items = []
        else:
            obj.aux_items = ArrayAuxItem.parse(buf)
        return obj

    def __eq__(self, other):
        return (self.creation_time == other.creation_time and
                self.author == other.author and
                self.no_of_lines == other.no_of_lines and
                self.no_of_chars == other.no_of_chars and
                self.no_of_marks == other.no_of_marks and
                self.misc_info == other.misc_info and
                self.aux_items == other.aux_items)

    def __ne__(self, other):
        return not self == other


# CONFERENCE

class ConfType(Bitstring4):
    rd_prot = property(*_create_bitstring_accessors(0))
    original = property(*_create_bitstring_accessors(1))
    secret = property(*_create_bitstring_accessors(2))
    letterbox = property(*_create_bitstring_accessors(3))

class ExtendedConfType(Bitstring8):
    def __init__(self, conf_type=None):
        if isinstance(conf_type, ConfType):
            conf_type = conf_type + [0]*(ExtendedConfType.LENGTH - ConfType.LENGTH)
        Bitstring8.__init__(self, conf_type)
    rd_prot = property(*_create_bitstring_accessors(0))
    original = property(*_create_bitstring_accessors(1))
    secret = property(*_create_bitstring_accessors(2))
    letterbox = property(*_create_bitstring_accessors(3))
    allow_anonymous = property(*_create_bitstring_accessors(4))
    forbid_secret = property(*_create_bitstring_accessors(5))
    reserved2 = property(*_create_bitstring_accessors(6))
    reserved3 = property(*_create_bitstring_accessors(7))

class AnyConfType(ExtendedConfType):
    """Alias for ExtendedConfType.
    """
    pass



class Conference(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.name = String.parse(buf)
        obj.type = ExtendedConfType.parse(buf)
        obj.creation_time = Time.parse(buf)
        obj.last_written = Time.parse(buf)
        obj.creator = PersNo.parse(buf)
        obj.presentation = TextNo.parse(buf)
        obj.supervisor = ConfNo.parse(buf)
        obj.permitted_submitters = ConfNo.parse(buf)
        obj.super_conf = ConfNo.parse(buf)
        obj.msg_of_day = TextNo.parse(buf)
        obj.nice = GarbNice.parse(buf)
        obj.keep_commented = GarbNice.parse(buf)
        obj.no_of_members = Int16.parse(buf)
        obj.first_local_no = LocalTextNo.parse(buf)
        obj.no_of_texts = Int32.parse(buf)
        obj.expire = GarbNice.parse(buf)
        obj.aux_items = ArrayAuxItem.parse(buf)
        return obj

    def __str__(self):
        return "<Conference %s>" % self.name
    
class UConference(object):
    def __init__(self, name=None, conf_type=None, highest_local_no=0, nice=0):
        if name is None:
            name = ""
        if conf_type is None:
            conf_type = ExtendedConfType()

        self.name = name
        self.type = conf_type
        self.highest_local_no = highest_local_no
        self.nice = nice

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.name = String.parse(buf)
        obj.type = ExtendedConfType.parse(buf)
        obj.highest_local_no = LocalTextNo.parse(buf)
        obj.nice = GarbNice.parse(buf)
        return obj

    def __eq__(self, other):
        return (self.name == other.name and
                self.type == other.type and
                self.highest_local_no == other.highest_local_no and
                self.nice == other.nice)

    def __repr__(self):
        return ("UConference(name={name!r}, conf_type={type!r}, "
                "highest_local_no={hln!r}, nice={nice!r})".format(
                name=self.name,
                type=self.type,
                hln=self.highest_local_no,
                nice=self.nice))

    def __str__(self):
        return "<UConference %s>" % self.name
    
# PERSON

class PrivBits(Bitstring16):
    wheel = property(*_create_bitstring_accessors(0))
    admin = property(*_create_bitstring_accessors(1))
    statistic = property(*_create_bitstring_accessors(2))
    create_pers = property(*_create_bitstring_accessors(3))
    create_conf = property(*_create_bitstring_accessors(4))
    change_name = property(*_create_bitstring_accessors(5))
    flg7 = property(*_create_bitstring_accessors(6))
    flg8 = property(*_create_bitstring_accessors(7))
    flg9 = property(*_create_bitstring_accessors(8))
    flg10 = property(*_create_bitstring_accessors(9))
    flg11 = property(*_create_bitstring_accessors(10))
    flg12 = property(*_create_bitstring_accessors(11))
    flg13 = property(*_create_bitstring_accessors(12))
    flg14 = property(*_create_bitstring_accessors(13))
    flg15 = property(*_create_bitstring_accessors(14))
    flg16 = property(*_create_bitstring_accessors(15))
    
class PersonalFlags(Bitstring8):
    unread_is_secret = property(*_create_bitstring_accessors(0))
    flg2 = property(*_create_bitstring_accessors(1))
    flg3 = property(*_create_bitstring_accessors(2))
    flg4 = property(*_create_bitstring_accessors(3))
    flg5 = property(*_create_bitstring_accessors(4))
    flg6 = property(*_create_bitstring_accessors(5))
    flg7 = property(*_create_bitstring_accessors(6))
    flg8 = property(*_create_bitstring_accessors(7))

class Person(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.username = String.parse(buf)
        obj.privileges = PrivBits.parse(buf)
        obj.flags = PersonalFlags.parse(buf)
        obj.last_login = Time.parse(buf)
        obj.user_area = TextNo.parse(buf)
        obj.total_time_present = Int32.parse(buf)
        obj.sessions = Int32.parse(buf)
        obj.created_lines = Int32.parse(buf)
        obj.created_bytes = Int32.parse(buf)
        obj.read_texts = Int32.parse(buf)
        obj.no_of_text_fetches = Int32.parse(buf)
        obj.created_persons = Int16.parse(buf)
        obj.created_confs = Int16.parse(buf)
        obj.first_created_local_no = Int32.parse(buf)
        obj.no_of_created_texts = Int32.parse(buf)
        obj.no_of_marks = Int16.parse(buf)
        obj.no_of_confs = Int16.parse(buf)
        return obj

# MEMBERSHIP

class MembershipType(Bitstring):
    LENGTH = 8
    invitation = property(*_create_bitstring_accessors(0))
    passive = property(*_create_bitstring_accessors(1))
    secret = property(*_create_bitstring_accessors(2))
    passive_message_invert = property(*_create_bitstring_accessors(3))
    reserved2 = property(*_create_bitstring_accessors(4))
    reserved3 = property(*_create_bitstring_accessors(5))
    reserved4 = property(*_create_bitstring_accessors(6))
    reserved5 = property(*_create_bitstring_accessors(7))

class Membership10(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.position = Int32.parse(buf)
        obj.last_time_read  = Time.parse(buf)
        obj.conference = ConfNo.parse(buf)
        obj.priority = Int8.parse(buf)
        obj.last_text_read = LocalTextNo.parse(buf)
        obj.read_texts = ArrayLocalTextNo.parse(buf)
        obj.added_by = PersNo.parse(buf)
        obj.added_at = Time.parse(buf)
        obj.type = MembershipType.parse(buf)
        return obj

class ReadRange(object):
    def __init__(self, first_read = 0, last_read = 0):
        self.first_read = first_read
        self.last_read = last_read
        
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.first_read = LocalTextNo.parse(buf)
        obj.last_read = LocalTextNo.parse(buf)
        return obj

    def __str__(self):
        return "<ReadRange %d-%d>" % (self.first_read, self.last_read)

    def to_string(self):
        return b"%d %d" % \
            (self.first_read,
             self.last_read)

class ArrayReadRange(Array):
    ELEMENT_CLASS = ReadRange

class Membership11(object):
    def __init__(self, position=0, last_time_read=None, conference=0, priority=0,
                 read_ranges=None, added_by=0, added_at=None, membership_type=None):
        if last_time_read is None:
            last_time_read = Time()
        if read_ranges is None:
            read_ranges = ArrayReadRange()
        if added_at is None:
            added_at = Time()
        if membership_type is None:
            membership_type = MembershipType()

        self.position = position
        self.last_time_read = last_time_read
        self.conference = conference
        self.priority = priority
        self.read_ranges = read_ranges
        self.added_by = added_by
        self.added_at = added_at
        self.type = membership_type

    def __repr__(self):
        return ("Membership11(position={pos!r}, last_time_read={ltr!r}, conference={conf!r}, "
             "priority={prio!r}, read_ranges={rr!r}, added_by={by!r}, added_at={at!r}, "
             "membership_type={type!r})".format(
                pos=self.position,
                ltr=self.last_time_read,
                conf=self.conference,
                prio=self.priority,
                rr=self.read_ranges,
                by=self.added_by,
                at=self.added_at,
                type=self.type))

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.position = Int32.parse(buf)
        obj.last_time_read  = Time.parse(buf)
        obj.conference = ConfNo.parse(buf)
        obj.priority = Int8.parse(buf)
        obj.read_ranges = ArrayReadRange.parse(buf)
        obj.added_by = PersNo.parse(buf)
        obj.added_at = Time.parse(buf)
        obj.type = MembershipType.parse(buf)
        return obj

    def __eq__(self, other):
        return (self.position == other.position and
                self.last_time_read == other.last_time_read and
                self.conference == other.conference and
                self.priority == other.priority and
                self.read_ranges == other.read_ranges and
                self.added_by == other.added_by and
                self.added_at == other.added_at and
                self.type == other.type)

Membership = Membership11

class Member(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.member  = PersNo.parse(buf)
        obj.added_by = PersNo.parse(buf)
        obj.added_at = Time.parse(buf)
        obj.type = MembershipType.parse(buf)
        return obj

# TEXT LIST

class TextList(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.first_local_no = LocalTextNo.parse(buf)
        obj.texts = ArrayTextNo.parse(buf)
        return obj

# TEXT MAPPING

class TextNumberPair(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.local_number = LocalTextNo.parse(buf)
        obj.global_number = TextNo.parse(buf)
        return obj

class ArrayTextNumberPair(Array):
    ELEMENT_CLASS = TextNumberPair

class TextMapping(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.range_begin = LocalTextNo.parse(buf) # Included in the range
        obj.range_end = LocalTextNo.parse(buf) # Not included in range (first after)
        obj.later_texts_exists = Bool.parse(buf)
        obj.block_type = Int32.parse(buf)

        obj.dict = {}
        obj.list = []

        if obj.block_type == 0:
            # Sparse
            obj.type_text = "sparse"
            obj.sparse_list = ArrayTextNumberPair.parse(buf)
            for tnp in obj.sparse_list:
                obj.dict[tnp.local_number] = tnp.global_number
                obj.list.append((tnp.local_number, tnp.global_number))
        elif obj.block_type == 1:
            # Dense
            obj.type_text = "dense"
            obj.dense_first = LocalTextNo.parse(buf)
            obj.dense_texts = ArrayInt32.parse(buf)
            local_number = obj.dense_first
            for global_number in obj.dense_texts:
                obj.dict[local_number] = global_number
                obj.list.append((local_number, global_number))
                local_number = local_number + 1
        else:
            raise ProtocolError
        return obj

    def __str__(self):
        if self.later_texts_exists:
            more = " (more exists)"
        else:
            more = ""
        return "<TextMapping (%s) %d...%d%s>" % (
            self.type_text,
            self.range_begin, self.range_end - 1 ,
            more)
# MARK

class Mark(object):
    def __init__(self, text_no=0, type=0):
        self.text_no = text_no
        self.type = type

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.text_no = TextNo.parse(buf)
        obj.type = Int8.parse(buf)
        return obj

    def __str__(self):
        return "<Mark %d (%d)>" % (self.text_no, self.type)

    def __eq__(self, other):
        return (self.text_no == other.text_no and
                self.type == other.type)
    
    def __ne__(self, other):
        return not self == other


# SERVER INFORMATION

class Info(object):
    def __init__(self):
        self.version = None
        self.conf_pres_conf = None
        self.pers_pres_conf = None
        self.motd_conf = None
        self.kom_news_conf = None
        self.motd_of_lyskom = None
        self.aux_item_list = []

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.version = Int32.parse(buf)
        obj.conf_pres_conf = ConfNo.parse(buf)
        obj.pers_pres_conf = ConfNo.parse(buf)
        obj.motd_conf = ConfNo.parse(buf)
        obj.kom_news_conf = ConfNo.parse(buf)
        obj.motd_of_lyskom = TextNo.parse(buf)
        obj.aux_item_list = ArrayAuxItem.parse(buf)
        return obj


class InfoOld(object):
    def __init__(self, info_old=None, version=0, conf_pres_conf=0, pers_pres_conf=0,
                 motd_conf=0, kom_news_conf=0, motd_of_lyskom=0):
        if info_old is None:
            self.version = version
            self.conf_pres_conf = conf_pres_conf
            self.pers_pres_conf = pers_pres_conf
            self.motd_conf = motd_conf
            self.kom_news_conf = kom_news_conf
            self.motd_of_lyskom = motd_of_lyskom
        else:
            self.version = info_old.version
            self.conf_pres_conf = info_old.conf_pres_conf
            self.pers_pres_conf = info_old.pers_pres_conf
            self.motd_conf = info_old.motd_conf
            self.kom_news_conf = info_old.kom_news_conf
            self.motd_of_lyskom = info_old.motd_of_lyskom

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.version = Int32.parse(buf)
        obj.conf_pres_conf = ConfNo.parse(buf)
        obj.pers_pres_conf = ConfNo.parse(buf)
        obj.motd_conf = ConfNo.parse(buf)
        obj.kom_news_conf = ConfNo.parse(buf)
        obj.motd_of_lyskom = TextNo.parse(buf)
        return obj

    def to_string(self):
        return b"%d %d %d %d %d %d" % (
            self.version,
            self.conf_pres_conf,
            self.pers_pres_conf,
            self.motd_conf,
            self.kom_news_conf,
            self.motd_of_lyskom)


class VersionInfo(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.protocol_version = Int32.parse(buf)
        obj.server_software = String.parse(buf)
        obj.software_version = String.parse(buf)
        return obj

    def __str__(self):
        return "<VersionInfo protocol %d by %s %s>" % \
               (self.protocol_version,
                self.server_software, self.software_version)

# New in protocol version 11
class StaticServerInfo(object): 
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.boot_time = Time.parse(buf)
        obj.save_time = Time.parse(buf)
        obj.db_status = String.parse(buf)
        obj.existing_texts = Int32.parse(buf)
        obj.highest_text_no = TextNo.parse(buf)
        obj.existing_confs = Int32.parse(buf)
        obj.existing_persons = Int32.parse(buf)
        obj.highest_conf_no = ConfNo.parse(buf)
        return obj

    def __str__(self):
        return "<StaticServerInfo>"

# SESSION INFORMATION

class SessionFlags(Bitstring):
    LENGTH = 8
    invisible = property(*_create_bitstring_accessors(0))
    user_active_used = property(*_create_bitstring_accessors(1))
    user_absent = property(*_create_bitstring_accessors(2))
    reserved3 = property(*_create_bitstring_accessors(3))
    reserved4 = property(*_create_bitstring_accessors(4))
    reserved5 = property(*_create_bitstring_accessors(5))
    reserved6 = property(*_create_bitstring_accessors(6))
    reserved7 = property(*_create_bitstring_accessors(7))

class DynamicSessionInfo(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.session = SessionNo.parse(buf)
        obj.person = PersNo.parse(buf)
        obj.working_conference = ConfNo.parse(buf)
        obj.idle_time = Int32.parse(buf)
        obj.flags = SessionFlags.parse(buf)
        obj.what_am_i_doing  = String.parse(buf)
        return obj

class StaticSessionInfo(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.username = String.parse(buf)
        obj.hostname = String.parse(buf)
        obj.ident_user = String.parse(buf)
        obj.bufection_time = Time.parse(buf)
        return obj

class SchedulingInfo(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.priority = Int16.parse(buf)
        obj.weight = Int16.parse(buf)
        return obj

class WhoInfo(object):
    def __init__(self, person=0, working_conference=0, session=0,
                 what_am_i_doing=None, username=None):
        if what_am_i_doing is None:
            what_am_i_doing = ""
        if username is None:
            username = ""
        self.person = person
        self.working_conference = working_conference
        self.session = session
        self.what_am_i_doing = what_am_i_doing
        self.username = username

    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.person = PersNo.parse(buf)
        obj.working_conference = ConfNo.parse(buf)
        obj.session = SessionNo.parse(buf)
        obj.what_am_i_doing  = String.parse(buf)
        obj.username = String.parse(buf)
        return obj

    def __eq__(self, other):
        return (self.person == other.person and
                self.working_conference == other.working_conference and
                self.session == other.session and 
                self.what_am_i_doing == other.what_am_i_doing and
                self.username == other.username)

    def __ne__(self, other):
        return not self == other
     
# STATISTICS

class StatsDescription(object):
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.what = ArrayString.parse(buf)
        obj.when = ArrayInt32.parse(buf)
        return obj
     
    def __str__(self):
        return "<StatsDescription>"

    def __eq__(self, other):
        return (self.what == other.what and
                self.when == other.when)

    def __ne__(self, other):
        return not self == other

class Stats(object):
    def __init__(self, average=0.0, ascent_rate=0.0, descent_rate=0.0):
        self.average = average
        self.ascent_rate = ascent_rate
        self.descent_rate = descent_rate
        
    @classmethod
    def parse(cls, buf):
        obj = cls()
        obj.average = Float.parse(buf)
        obj.ascent_rate = Float.parse(buf)
        obj.descent_rate = Float.parse(buf)
        return obj

    def __str__(self):
        return "<Stats %f + %f - %f>" % (self.average,
                                         self.ascent_rate,
                                         self.descent_rate)

    def __eq__(self, other):
        return (self.average == other.average and
                self.ascent_rate == other.ascent_rate and
                self.descent_rate == other.descent_rate)

    def __ne__(self, other):
        return not self == other


class ArrayMark(Array):
    ELEMENT_CLASS = Mark

class ArrayMember(Array):
    ELEMENT_CLASS = Member

class ArrayMembership11(Array):
    ELEMENT_CLASS = Membership11

class ArrayMembership10(Array):
    ELEMENT_CLASS = Membership10

class ArrayStats(Array):
    ELEMENT_CLASS = Stats

class ArrayConfNo(Array):
    ELEMENT_CLASS = ConfNo

class ArrayConfZInfo(Array):
    ELEMENT_CLASS = ConfZInfo

class ArrayDynamicSessionInfo(Array):
    ELEMENT_CLASS = DynamicSessionInfo



# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.

from __future__ import absolute_import
import base64
import functools
import json
import six

import errno
import socket
from . import mimeparse

from . import komauxitems, utils, requests
from .connection import Connection
from .cachedconnection import Client, CachingPersonClient
from .stats import stats

from .datatypes import (
    AuxItemInput,
    ConfType,
    CookedMiscInfo,
    MIC_COMMENT,
    MIC_FOOTNOTE,
    MICommentTo,
    MIR_BCC,
    MIR_CC,
    MIR_TO,
    MIRecipient,
    MembershipType,
    PersonalFlags,
    first_aux_items_with_tag)

class KomSessionException(Exception): pass
class KomSessionNotConnected(KomSessionException): pass
class KomSessionError(KomSessionException): pass
class AmbiguousName(KomSessionError): pass
class NameNotFound(KomSessionError): pass
class NoRecipients(KomSessionError): pass


MIRecipient_str_to_type = { 'to': MIR_TO,
                            'cc': MIR_CC,
                            'bcc': MIR_BCC }
        
MICommentTo_str_to_type = { 'comment': MIC_COMMENT,
                            'footnote': MIC_FOOTNOTE }


def create_client(host, port, user):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    conn = Connection(s, user)
    client = Client(conn)
    return CachingPersonClient(client)


def check_connection(f):
    @functools.wraps(f)
    def decorated(komsession, *args, **kwargs):
        if not komsession.is_connected():
            raise KomSessionNotConnected()
        try:
            return f(komsession, *args, **kwargs)
        except socket.error as serr:
            if serr.errno in (errno.EPIPE, errno.ECONNRESET, errno.ENOTCONN, errno.ETIMEDOUT):
                # If we got an error that indicates that the
                # connection has failed, then close and raise.
                komsession.close()
                raise KomSessionNotConnected(serr)
            else:
                raise KomSessionException(serr)

    return decorated


# Idea: rename KomSession to KomClient?
class KomSession(object):
    """ A LysKom session.

    Should handle either unicode strings or utf-8 encoded strings. (FIXME)

    TODO[Python3]: Only handle (unicode) strings, not bytes, in the
    external interfaces for things that are real strings. ??? (Or only
    bytes? Seems inconvient at this level.)

    """
    def __init__(self, client_factory=create_client):
        # TODO: We actually require the API of a
        # CachingPersonClient. We should enhance the Connection
        # class and make CachingPersonClient have the same API as
        # Connection.
        self._client_factory = client_factory
        self._client = None
        self._session_no = None
        self._client_name = None
        self._client_version = None

    def connect(self, host, port, username, hostname, client_name, client_version):
        assert not self.is_connected() # todo: raise better exception
        # decode if not already unicode (assuming utf-8)
        if isinstance(client_name, six.binary_type):
            client_name = client_name.decode('utf-8')
        if isinstance(client_version, six.binary_type):
            client_version = client_version.decode('utf-8')

        self._client = self._client_factory(host, port, user=username + "%" + hostname)

        # todo: we shouldn't require client name/version. specify in
        # constructor instead (because I don't think it should be
        # possible to change after connecting) - but send the request
        # here (if they are set).
        self._client.request(requests.ReqSetClientVersion(client_name, client_version))
        self._client_name = client_name
        self._client_version = client_version

        self._session_no = self.who_am_i()
        self._client.request(requests.ReqSetConnectionTimeFormat(use_utc=1))
    
    def is_connected(self):
        return self._client is not None
    
    def close(self):
        """Immediately close the connection, without sending a Disconnect request.
        """
        try:
            if self._client is not None:
                self._client.close()
        finally:
            self._client = None
            self._client_name = None
            self._client_version = None
            self._session_no = None

    @check_connection
    def disconnect(self, session_no=0):
        """Session number 0 means this session (a logged in user can
        disconnect its other sessions).
        """
        self._client.request(requests.ReqDisconnect(session_no))

        # Check if we disconnected our own session or not (you can
        # disconnect another LysKOM session that the logged in user is
        # a supervisor of).
        if session_no == 0 or session_no == self._session_no:
            self.close()

    @check_connection
    def login(self, pers_no, password):
        if isinstance(password, six.binary_type):
            password = password.decode('utf-8')
        pers_no = int(pers_no)
        self._client.login(pers_no, password)
        person_stat = self._client.request(requests.ReqGetPersonStat(pers_no))
        return KomPerson(pers_no, person_stat)

    @check_connection
    def logout(self):
        self._client.logout()

    @check_connection
    def get_person_no(self):
        return self._client.get_person_no()

    @check_connection
    def who_am_i(self):
        return self._client.request(requests.ReqWhoAmI())

    @check_connection
    def user_is_active(self):
        self._client.request(requests.ReqUserActive())

    @check_connection
    def is_logged_in(self):
        return self._client.is_logged_in()

    @check_connection
    def change_conference(self, conf_no):
        self._client.change_conference(conf_no)
        
    @check_connection
    def create_person(self, name, passwd):
        # decode if not already unicode (assuming utf-8)
        if isinstance(name, six.binary_type):
            name = name.decode('utf-8')
        if isinstance(passwd, six.binary_type):
            passwd = passwd.decode('utf-8')

        flags = PersonalFlags()
        aux_items = []
        pers_no = self._client.request(
            requests.ReqCreatePerson(name, passwd, flags, aux_items))
        stats.set('komsession.persons.created.last', 1, agg='sum')
        return KomPerson(pers_no)

    @check_connection
    def create_conference(self, name, aux_items=None):
        # decode if not already unicode (assuming utf-8)
        if isinstance(name, six.binary_type):
            name = name.decode('utf-8')

        conf_type = ConfType()
        if aux_items is None:
            aux_items = []
        conf_no = self._client.request(
            requests.ReqCreateConf(name.encode('latin1'), conf_type, aux_items))
        stats.set('komsession.conferences.created.last', 1, agg='sum')
        return conf_no

    @check_connection
    def lookup_name(self, name, want_pers, want_confs):
        if isinstance(name, six.binary_type):
            name = name.decode('utf-8')
        return self._client.lookup_name(name, want_pers, want_confs)

    def lookup_name_exact(self, name, want_pers, want_confs):
        matches = self.lookup_name(name, want_pers, want_confs)
        return self._exact_lookup_match(name, matches)

    @check_connection
    def re_lookup_name(self, regexp, want_pers, want_confs):
        # The LysKOM server is always case sensitive, and it's kom.py
        # that tries to create a case-insensitive regexp. Doesn't seem
        # to work that well.
        return self._client.regexp_lookup(regexp, want_pers, want_confs, case_sensitive=1)

    def re_lookup_name_exact(self, regexp, want_pers, want_confs):
        matches = self.re_lookup_name(regexp, want_pers, want_confs)
        return self._exact_lookup_match(regexp, matches)

    @staticmethod
    def _exact_lookup_match(lookup, matches):
        if len(matches) == 0:
            raise NameNotFound("recipient not found: %s" % lookup)
        elif len(matches) != 1:
            raise AmbiguousName("ambiguous recipient: %s" % lookup)
        return matches[0][0]

    @check_connection
    def get_text_stat(self, text_no):
        return self._client.textstats[text_no]
    
    @check_connection
    def add_membership(self, pers_no, conf_no, priority, where):
        mtype = MembershipType()
        self._client.request(requests.ReqAddMember(conf_no, pers_no, priority, where, mtype))
    
    @check_connection
    def delete_membership(self, pers_no, conf_no):
        self._client.request(requests.ReqSubMember(conf_no, pers_no))

    @check_connection
    def get_membership(self, pers_no, conf_no):
        membership = self._client.get_membership(pers_no, conf_no, want_read_ranges=False)
        return KomMembership(pers_no, membership)

    @check_connection
    def get_membership_unread(self, pers_no, conf_no):
        membership = self._client.get_membership(pers_no, conf_no, want_read_ranges=True)
        unread_texts = self._client.get_unread_texts_from_membership(membership)
        return KomMembershipUnread(pers_no, conf_no, len(unread_texts), unread_texts)

    @check_connection
    def get_memberships(self, pers_no, first, no_of_confs, unread=False, passive=False):
        if unread:
            # RegGetUnreadConfs never returns passive memberships so
            # that combination is not valid.
            assert passive == False
            conf_nos = self._client.request(requests.ReqGetUnreadConfs(pers_no))
            # This may return conferences that don't have any unread
            # texts in them. We have to live with this, because we
            # don't want to get the unread texts in this case. It's
            # possible that we need to change this, which means that
            # unread=True may be a slower call.
            memberships = [ self.get_membership(pers_no, conf_no) for conf_no in conf_nos ]
            has_more = False
        else:
            ms_list = self._client.get_memberships(pers_no, first, no_of_confs,
                                                 want_read_ranges=False)
            
            # We need to check if there are more memberships to get
            # before we filter out the passive memberships.
            if len(ms_list) < no_of_confs:
                has_more = False
            else:
                has_more = True
            
            memberships = []
            for membership in ms_list:
                if (not passive) and membership.type.passive:
                    continue
                memberships.append(KomMembership(pers_no, membership))
        
        return memberships, has_more

    @check_connection
    def get_membership_unreads(self, pers_no):
        conf_nos = self._client.request(requests.ReqGetUnreadConfs(pers_no))
        memberships = [ self.get_membership_unread(pers_no, conf_no)
                        for conf_no in conf_nos ]
        return [ m for m in memberships if m.no_of_unread > 0 ]
    
    @check_connection
    def get_conf_name(self, conf_no):
        return self._client.conf_name(conf_no)
    
    @check_connection
    def get_conference(self, conf_no, micro=True):
        conf_no = int(conf_no)
        if micro:
            return KomUConference(conf_no, self._client.uconferences[conf_no])
        else:
            return KomConference(conf_no, self._client.conferences[conf_no])

    @check_connection
    def get_text(self, text_no):
        text_stat = self.get_text_stat(text_no)
        text = self._client.request(requests.ReqGetText(text_no))
        return KomText(text_no=text_no, text=text, text_stat=text_stat)

    # TODO: offset/start number, so we can paginate. we probably need
    # to return the local text number for that.
    @check_connection
    def get_last_texts(self, conf_no, no_of_texts, offset=0, full_text=False):
        """Get the {no_of_texts} last texts in conference {conf_no},
        starting from {offset}.
        """
        #local_no_ceiling = 0 # means the higest numbered texts (i.e. the last)
        text_mapping = self._client.request(
            requests.ReqLocalToGlobalReverse(conf_no, 0, no_of_texts))
        texts = [ KomText(text_no=m[1], text=None, text_stat=self.get_text_stat(m[1]))
                  for m in text_mapping.list if m[1] != 0 ]
        texts.reverse()
        return texts

    @check_connection
    def create_text(self, subject, body, content_type, content_encoding=None,
                    recipient_list=None, comment_to_list=None):
        # decode if not already unicode (assuming utf-8)
        if isinstance(subject, six.binary_type):
            subject = subject.decode('utf-8')
        if isinstance(body, six.binary_type):
            body = body.decode('utf-8')
        if isinstance(content_type, six.binary_type):
            content_type = content_type.decode('utf-8')

        if content_encoding is not None:
            if content_encoding == "base64":
                body = base64.b64decode(body)
            else:
               raise ValueError("Invalid content_encoding: %s", content_encoding)

        # wtf are we doing here?
        mime_type, _ = utils.parse_content_type(content_type)
        content_type = utils.mime_type_tuple_to_str(mime_type)
        mime_type = mimeparse.parse_mime_type(content_type)

        if mime_type[0] == 'text':
            # We hard code utf-8 because it is The Correct Encoding. :)
            mime_type[2]['charset'] = 'utf-8'
            fulltext = subject + "\n" + body
            fulltext = fulltext.encode('utf-8')
        elif mime_type[0] == 'x-kom' and mime_type[1] == 'user-area':
            # Charset doesn't seem to be specified for user areas, but
            # in reality they contain Latin 1 text.
            fulltext = body.encode('latin-1')
        elif mime_type[0] == 'image':
            # We handle images in the same way as AndroKOM. That means
            # that the subject is encoded with latin-1, and the stored
            # text is latin-1-subject + "\n" + binary-body. Content
            # type is "image/jpeg; name=image:<???>". Note that there
            # is no information regarding how subjects are encoded.
            #
            # TODO: What do we do if we can't encode the subject with
            # latin-1?
            fulltext = subject.encode('latin-1') + b"\n" + body
        else:
            raise KomSessionError("Unhandled content type: %s" % (mime_type,))


        misc_info = CookedMiscInfo()

        if recipient_list is not None:
            for r in recipient_list:
                misc_info.recipient_list.append(
                    MIRecipient(type=MIRecipient_str_to_type[r['type']],
                                recpt=r['recpt']['conf_no']))

        if comment_to_list is not None:
            for ct in comment_to_list:
                misc_info.comment_to_list.append(
                    MICommentTo(type=MICommentTo_str_to_type[ct['type']],
                                text_no=ct['text_no']))

        aux_items = []

        # We need to make sure all aux items are encoded.
        creating_software = "%s %s" % (self._client_name, self._client_version)
        aux_items.append(AuxItemInput(tag=komauxitems.AI_CREATING_SOFTWARE,
                                      data=creating_software.encode('utf-8')))
        aux_items.append(AuxItemInput(tag=komauxitems.AI_CONTENT_TYPE,
                                 data=utils.mime_type_tuple_to_str(mime_type).encode('utf-8')))

        text_no = self._client.request(
            requests.ReqCreateText(fulltext, misc_info, aux_items))
        stats.set('komsession.texts.created.last', 1, agg='sum')
        return text_no

    @check_connection
    def mark_as_read(self, text_no):
        text_stat = self.get_text_stat(text_no)
        for mi in text_stat.misc_info.recipient_list:
            self._client.mark_as_read_local(mi.recpt, mi.loc_no)

    @check_connection
    def mark_as_unread(self, text_no):
        text_stat = self.get_text_stat(text_no)
        for mi in text_stat.misc_info.recipient_list:
            self._client.mark_as_unread_local(mi.recpt, mi.loc_no)

    @check_connection
    def set_unread(self, conf_no, no_of_unread):
        self._client.request(requests.ReqSetUnread(conf_no, no_of_unread))

    @check_connection
    def get_marks(self):
        return self._client.request(requests.ReqGetMarks())

    @check_connection
    def mark_text(self, text_no, mark_type):
        self._client.mark_text(text_no, mark_type)

    @check_connection
    def unmark_text(self, text_no):
        self._client.unmark_text(text_no)

    @check_connection
    def get_user_area_block(self, pers_no, block_name, json_decode=True):
        """Get the block with the given block name from the user area
        for the given person. If there is no user area for the person,
        or if there is no block with the given name, None will be
        returned.

        If json_decode is True (default), the stored block will be
        passed to json.loads() before it is returned.

        If json_decode is False, then the block will be returned as a
        string.
        """
        person_stat = self._client.request(requests.ReqGetPersonStat(pers_no))

        if person_stat.user_area == 0:
            # No user area
            return None

        # TODO: don't use external get_text method here - it should decode the body,
        # but we don't want to do that.
        text = self.get_text(person_stat.user_area)
        if text.content_type != 'x-kom/user-area':
            raise KomSessionError(
                "Unknown content type for user area text: %s" % (text.content_type,))

        blocks = utils.decode_user_area(text.body.encode('latin1')) #HACK
        block = blocks.get(block_name, None)
        if block is not None and json_decode:
            block = json.loads(block.decode('latin1')) #HACK

        return block

    @check_connection
    def set_user_area_block(self, pers_no, block_name, block, json_encode=True):
        """Set the block with the given block name in the user area
        for the given person. Will create a new text and set as the
        new user area for the person. If there already is a block with
        the given name, it will be over-written. The other blocks in
        the user area will copied to the new user area.

        If json_encode is True (default), the block will be passed to
        json.dumps() before it is saved.

        If json_encode is False, then the block should be a string
        that can be hollerith encoded.
        """
        person_stat = self._client.request(requests.ReqGetPersonStat(pers_no))

        if person_stat.user_area == 0:
            # No existing user area, initiate a new dictionary of
            # blocks.
            blocks = dict()
        else:
            # TODO: don't use external get_text method here - it
            # should decode the body, but we don't want to do that.
            user_area = self.get_text(person_stat.user_area)
            if user_area.content_type != 'x-kom/user-area':
                raise KomSessionError(
                    "Unknown content type for user area text: %s" % (user_area.content_type,))

            blocks = utils.decode_user_area(user_area.body.encode('latin1')) # HACK

        if json_encode:
            blocks[block_name] = json.dumps(block).encode('latin1') # HACK
        else:
            blocks[block_name] = block

        new_user_area_text_no = self.create_text(
            subject=None,
            body=utils.encode_user_area(blocks),
            content_type='x-kom/user-area')
        self._client.request(requests.ReqSetUserArea(pers_no, new_user_area_text_no))
        # TODO: Should it remove the old user area?


class KomPerson(object):
    def __init__(self, pers_no, person_stat=None):
        self.pers_no = pers_no

        if person_stat is None:
            #self.username = None
            #self.privileges = None
            #self.flags = None
            #self.last_login = None
            self.user_area = None
            #self.total_time_present = None
            #self.sessions = None
            #self.created_lines = None
            #self.created_bytes = None
            #self.read_texts = None
            #self.no_of_text_fetches = None
            #self.user_area = None
            #self.created_persons = None
            #self.created_confs = None
            #self.first_created_local_no = None
            #self.no_of_created_texts = None
            #self.no_of_marks = None
            #self.no_of_cons = None
        else:
            self.user_area = person_stat.user_area


class KomMembership(object):
    def __init__(self, pers_no, membership):
        self.pers_no = pers_no
        self.position = membership.position
        self.last_time_read = membership.last_time_read
        self.conference = membership.conference
        self.priority = membership.priority
        self.added_by = membership.added_by
        self.added_at = membership.added_at
        self.type = membership.type


class KomMembershipUnread(object):
    def __init__(self, pers_no, conf_no, no_of_unread, unread_texts):
        self.pers_no = pers_no
        self.conf_no = conf_no
        self.no_of_unread = no_of_unread
        self.unread_texts = unread_texts


class KomConference(object):
    def __init__(self, conf_no=None, conf=None):
        self.conf_no = conf_no

        if conf is not None:
            self.name = conf.name.decode('latin1')
            self.type = conf.type
            self.creation_time = conf.creation_time
            self.last_written = conf.last_written
            self.creator = conf.creator
            self.presentation = conf.presentation
            self.supervisor = conf.supervisor
            self.permitted_submitters = conf.permitted_submitters
            self.super_conf = conf.super_conf
            self.msg_of_day = conf.msg_of_day
            self.nice = conf.nice
            self.keep_commented = conf.keep_commented
            self.no_of_members = conf.no_of_members
            self.first_local_no = conf.first_local_no
            self.no_of_texts = conf.no_of_texts
            self.expire = conf.expire
            self.aux_items = conf.aux_items


class KomUConference(object):
    """U stands for micro"""
    def __init__(self, conf_no=None, uconf=None):
        self.conf_no = conf_no

        if uconf is not None:
            self.name = uconf.name.decode('latin1')
            self.type = uconf.type
            self.highest_local_no = uconf.highest_local_no
            self.nice = uconf.nice


class KomText(object):
    def __init__(self, text_no=None, text=None, text_stat=None):
        self.text_no = text_no

        if text_stat is None:
            self.content_type = None
            self.creation_time = None
            self.author = None
            self.no_of_marks = 0
            self.recipient_list = None
            self.comment_to_list = None
            self.comment_in_list = None
            self.subject = None
            self.body = None
            self.aux_items = None
        else:
            mime_type, encoding = utils.parse_content_type(
                KomText._get_content_type_from_text_stat(text_stat))
            self.content_type = utils.mime_type_tuple_to_str(mime_type)

            self.creation_time = text_stat.creation_time
            self.author = text_stat.author
            self.no_of_marks = text_stat.no_of_marks
            self.recipient_list = text_stat.misc_info.recipient_list
            self.comment_to_list = text_stat.misc_info.comment_to_list
            self.comment_in_list = text_stat.misc_info.comment_in_list
            self.aux_items = text_stat.aux_items
            self.subject, self.body = KomText._decode_text(text, mime_type, encoding)

    @staticmethod
    def _decode_text(text, mime_type, encoding):
        if text is None:
            return None, None

        subject = None
        body = None

        # text_stat is required for this
        if mime_type[0] == "x-kom" and mime_type[1] == "user-area":
            body = utils.decode_text(text, encoding)
        else:
            # If a text has no linefeeds, it only has a body
            if text.find(b'\n') == -1:
                subject = "" # Should probably be None instead?
                rawbody = text
            else:
                rawsubject, rawbody = text.split(b'\n', 1)
                # TODO: should we always decode the subject?
                subject = utils.decode_text(rawsubject, encoding)

            if mime_type[0] == 'text':
                # Only decode body if media type is text, and not
                # an image, for example.  Also, if the subject is
                # empty, everything becomes the subject, which
                # will get decoded.  Figure out how to handle all
                # this. Assume empty subject means everything in
                # body?
                body = utils.decode_text(rawbody, encoding)
            else:
                body = rawbody

        return subject, body

    @staticmethod
    def _get_content_type_from_text_stat(text_stat):
        try:
            contenttype = first_aux_items_with_tag(
                text_stat.aux_items, komauxitems.AI_CONTENT_TYPE).data.decode('latin1')
        except AttributeError:
            contenttype = 'text/plain'
        return contenttype



# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.


from __future__ import absolute_import
import six

WHITESPACE = bytearray(b" \t\r\n")
DIGITS = bytearray(b"01234567890")
FLOAT_CHARS = DIGITS + bytearray(b"eE.-+")

ORD_0 = ord("0")

MAX_TEXT_SIZE = int(2**31-1)


def read_first_non_ws(buf):
    """Skip whitespace and return first non-ws character"""
    c = buf.receive_char()
    while c in WHITESPACE:
        c = buf.receive_char()
    return c

def read_int_and_next(buf):
    """Get an integer and next character from the receive buffer."""
    c = read_first_non_ws(buf)
    n = 0
    while c in DIGITS:
        n = n * 10 + (ord(c) - ORD_0)
        c = buf.receive_char()
    return (n, c)

def read_int(buf):
    """Get an integer from the receive buffer (discard next character)"""
    (c, n) = read_int_and_next(buf)
    return c

def read_float(buf):
    # Get a float from the receive buffer (discard next character)
    c = read_first_non_ws(buf)
    digs = []
    while c in FLOAT_CHARS:
        digs.append(c)
        c = buf.receive_char()
    return float(b"".join(digs))


def to_hstring(s):
    """To hollerith byte string
    """
    assert isinstance(s, six.binary_type)
    return b"%dH%s" % (len(s), s)



# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.

from __future__ import absolute_import
from __future__ import print_function
import logging

from six.moves import range

from . import requests
from .asyncmsg import AsyncMessages, async_dict
from .errors import NotMember, NoSuchLocalText, UnimplementedAsync
from .stats import stats


logger = logging.getLogger(__name__)


class Client(object):
    def __init__(self, conn):
        self._conn = conn
        self._ok_queue = {}
        self._error_queue = {}
        self._async_handler_func = None

    def close(self):
        self._conn.close()

    def request(self, request):
        """
        Send an request and return the response.
        """
        logger.debug("sending request: %s" % (request,))
        ref_no = self._conn.send_request(request)
        resp = self._wait_and_dequeue(ref_no)
        logger.debug("returning response for ref_no: %s" % (ref_no, ))
        return resp

    def set_async_handler(self, handler_func):
        """Set the async handler function.
        
        @param handler_func Function that will be called when an async
        message is received. Will receive the async message as
        argument. If handler_func is None, there will be no handling
        of async messages.
        """
        self._async_handler_func = handler_func

    def _wait_and_dequeue(self, ref_no):
        """Wait for a request to be answered, return response or raise
        error.
        """
        logger.debug("waiting for  response ref_no: %s" % (ref_no,))
        while ref_no not in self._ok_queue and \
              ref_no not in self._error_queue:
            self._read_response()

        if ref_no in self._ok_queue:
            resp = self._ok_queue[ref_no]
            logger.debug("got response %s ref_no: %s" % (resp, ref_no))
            del self._ok_queue[ref_no]
            return resp
        elif ref_no in self._error_queue:
            error = self._error_queue[ref_no]
            logger.debug("got error %s ref_no: %s" % (error, ref_no))
            del self._error_queue[ref_no]
            raise error
        else:
            raise RuntimeError("Got unknown ref-no: %r" % (ref_no,))

    def _read_response(self):
        ref_no, resp, error = self._conn.read_response()
        logger.debug("read response for ref_no: %s" % (ref_no,))
        if ref_no is None:
            # async message
            self._handle_async_message(resp)
        elif error is not None:
            # error reply
            self._error_queue[ref_no] = error
        else:
            # ok reply - resp can be None
            self._ok_queue[ref_no] = resp

    def _handle_async_message(self, msg):
        if self._async_handler_func is not None:
            self._async_handler_func(msg)



#
# CLASS for a connection with...
# * Caches for:
#   - UConference
#   - Conference
#   - Person
#   - TextStat 
#   - Subjects
#   No negative caching. No time-outs.
#   Some automatic invalidation (if accept-async called appropriately).
#
# * Lookup function (conference/person name -> numbers)
# * Helper function get_unread_texts to get a list of local and global
#   numbers of all unread text in a conference for a person

class CachingClient(object):
    def __init__(self, client):
        self._client = client

        # Caches
        #
        # TODO: Instead of exposing these dictionary like cache
        # objects, we could override the request method in the
        # Connection class, and call the cache objects if the request
        # type matches the cache. Then we could have a the same
        # interface as Connection, which makes it easier to use and
        # test (fewer methods). IMPORTANT: It would however make it
        # less clear that we might get a cached response and that
        # could be dangerous. Sometime it is okay with cached
        # responses, and sometimes it is not. How can we make it
        # possible to force no cached?
        self.uconferences = Cache(self._fetch_uconference, "UConference")
        self.conferences = Cache(self._fetch_conference, "Conference")
        self.persons = Cache(self._fetch_person, "Person")
        self.textstats = Cache(self._fetch_textstat, "TextStat")

        self._async_handlers = {}
        self._client.set_async_handler(self._handle_async_message)

        # Setup up async handlers for invalidating cache entries. Skip
        # sending accept-async until the last call.
        self._add_async_handler(AsyncMessages.NEW_NAME, self._cah_new_name)
        self._add_async_handler(AsyncMessages.LEAVE_CONF, self._cah_leave_conf)
        self._add_async_handler(AsyncMessages.DELETED_TEXT, self._cah_deleted_text)
        self._add_async_handler(AsyncMessages.NEW_TEXT, self._cah_new_text)
        self._add_async_handler(AsyncMessages.NEW_RECIPIENT, self._cah_new_recipient)
        self._add_async_handler(AsyncMessages.SUB_RECIPIENT, self._cah_sub_recipient)
        self._add_async_handler(AsyncMessages.NEW_MEMBERSHIP, self._cah_new_membership)
        self.request(requests.ReqAcceptAsync(list(self._async_handlers.keys())))


    def close(self):
        self._client.close()

    def request(self, request):
        return self._client.request(request)


    # Async handling

    def _handle_async_message(self, msg):
        if msg.MSG_NO in self._async_handlers:
            for handler in self._async_handlers[msg.MSG_NO]:
                handler(msg)

    def _add_async_handler(self, msg_no, handler):
        """Register a handler for a type of async message.

        @param msg_no Type of async message.

        @param handler Function that should be called when an async
        message of the specified type is received.
        """
        if msg_no not in async_dict:
            raise UnimplementedAsync
        if msg_no in self._async_handlers:
            self._async_handlers[msg_no].append(handler)
        else:
            self._async_handlers[msg_no] = [handler]

    def register_async_handler(self, msg_no, handler, skip_accept_async=False):
        """Add an async handler and tell the LysKOM sever to
        start sending async messages of that type.
        
        @param skip_accept_async Don't send an AcceptAsync request to
        the LysKOM server now. This is an optimization feature. The
        protocol request registers all async message types at once, so
        this is useful to be able to only have to send one
        request. First register all but the last async handlers with
        skip_accept_async=True, and then register the last one with
        skip_accept_async=False to send the request.
        """
        self._add_async_handler(msg_no, handler)
        if not skip_accept_async:
            self.request(requests.ReqAcceptAsync(list(self._async_handlers.keys())))


    # Handlers for asynchronous messages (internal use) FIXME: Most of
    # these handlers could do more clever things than just
    # invalidating.

    def _cah_new_name(self, msg):
        # A new name makes uconferences[].name invalid
        self.uconferences.invalidate(msg.conf_no)
        # A new name makes conferences[].name invalid
        self.conferences.invalidate(msg.conf_no)

    def _cah_leave_conf(self, msg):
        # Leaving a conference makes conferences[].no_of_members invalid
        self.conferences.invalidate(msg.conf_no)

    def _cah_deleted_text(self, msg):
        # Deletion of a text makes conferences[].no_of_texts invalid
        ts = msg.text_stat
        for rcpt in ts.misc_info.recipient_list:
            self.conferences.invalidate(rcpt.recpt)
            
    def _cah_new_text(self, msg):
        # A new text. conferences[].no_of_texts and
        # uconferences[].highest_local_no is invalid. Also invalidates
        # the textstats for the commented texts.
        for rcpt in msg.text_stat.misc_info.recipient_list:
            self.conferences.invalidate(rcpt.recpt)
            self.uconferences.invalidate(rcpt.recpt)
        for ct in msg.text_stat.misc_info.comment_to_list:
            self.textstats.invalidate(ct.text_no)
        # FIXME: A new text makes persons[author].no_of_created_texts invalid

    def _cah_new_recipient(self, msg):
        # Just like a new text; conferences[].no_of_texts and
        # uconferences[].highest_local_no gets invalid. 
        self.conferences.invalidate(msg.conf_no)
        self.uconferences.invalidate(msg.conf_no)
        # textstats.misc_info_recipient_list gets invalid as well.
        self.textstats.invalidate(msg.text_no)

    def _cah_sub_recipient(self, msg):
        # Invalid conferences[].no_of_texts
        self.conferences.invalidate(msg.conf_no)
        # textstats.misc_info_recipient_list gets invalid as well.
        self.textstats.invalidate(msg.text_no)

    def _cah_new_membership(self, msg):
        # Joining a conference makes conferences[].no_of_members invalid
        self.conferences.invalidate(msg.conf_no)


    # Fetching functions (internal use)
    def _fetch_uconference(self, no):
        return self.request(requests.ReqGetUconfStat(no))

    def _fetch_conference(self, no):
        return self.request(requests.ReqGetConfStat(no))

    def _fetch_person(self, no):
        return self.request(requests.ReqGetPersonStat(no))

    def _fetch_textstat(self, no):
        return self.request(requests.ReqGetTextStat(no))


    # Report cache usage
    def report_cache_usage(self):
        self.uconferences.report()
        self.conferences.report()
        self.persons.report()
        self.textstats.report()

    # Common operation: get name of conference (via uconference)
    def conf_name(self, conf_no, default = "", include_no = 0):
        try:
            conf_name = self.uconferences[conf_no].name.decode('latin1')
            if include_no:
                return "%s (#%d)" % (conf_name, conf_no)
            else:
                return conf_name
        except Exception:
            if default.find("%d") != -1:
                return default % conf_no
            else:
                return default

    # Lookup function (name -> (list of tuples(no, name))
    # Special case: "#number" is not looked up
    def lookup_name(self, name, want_pers, want_confs):
        if name[:1] == "#":
            # Numerical case
            try:
                no = int(name[1:]) # Exception if not int
                type = self.uconferences[no].type # Exception if not found
                name = self.uconferences[no].name.decode('latin1')
                if (want_pers and type.letterbox) or \
                   (want_confs and (not type.letterbox)):
                    return [(no, name)]
                else:
                    return []
            except Exception:
                return []
        else:
            # Alphabetical case
            matches = self.request(
                requests.ReqLookupZName(
                    name,
                    want_pers=want_pers,
                    want_confs=want_confs))
            return [(x.conf_no, x.name.decode('latin1')) for x in matches]

    def regexp_lookup(self, regexp, want_pers, want_confs,
                      case_sensitive=0):
        """Lookup name using regular expression"""
        if regexp.startswith("#"):
            return self.lookup_name(regexp, want_pers, want_confs)
        
        if not case_sensitive:
            regexp = self._case_insensitive_regexp(regexp)

        matches = self.request(
            requests.ReqReZLookup(
                regexp,
                want_persons=want_pers,
                want_confs=want_confs))
        return [(x.conf_no, x.name.decode('latin1')) for x in matches]

    def _case_insensitive_regexp(self, regexp):
        """Make regular expression case insensitive"""
        result = ""
        # FIXME: Cache collate_table
        collate_table = self.request(requests.ReqGetCollateTable())
        inside_brackets = 0
        for c in regexp:
            if c == "[":
                inside_brackets = 1

            if inside_brackets:
                eqv_chars = c
            else:
                eqv_chars = self._equivalent_chars(c, collate_table)
                
            if len(eqv_chars) > 1:
                result += "[%s]" % eqv_chars
            else:
                result += eqv_chars

            if c == "]":
                inside_brackets = 0

        return result

    def _equivalent_chars(self, c, collate_table):
        """Find all chars equivalent to c in collate table"""
        c_ord = ord(c)
        if c_ord >= len(collate_table):
            return c

        result = ""
        norm_char = collate_table[c_ord]
        next_index = 0
        while 1:
            next_index = collate_table.find(norm_char, next_index)
            if next_index == -1:
                break
            result += chr(next_index)
            next_index += 1

        return result

    def read_ranges_to_gaps_and_last(self, read_ranges):
        """Return all texts excluded from read_ranges.
        
        @return: Returns a 2-tuple of a list and the first possibly
        unread text number after the last read range. The text number
        could be larger than the highest local number, if we have read
        everything in the conference. The list contains a 2-tuples for
        each gap in the read ranges, where each tuple is the first
        unread text in the gap and the length of the gap.
        """
        gaps = []
        last = 1
        for read_range in read_ranges:
            gap = read_range.first_read - last
            if gap > 0:
                gaps.append((last, gap))
            last = read_range.last_read + 1
        return gaps, last

    def get_unread_texts_from_membership(self, membership):
        unread = []
        
        more_to_fetch = 1
        gaps, last = self.read_ranges_to_gaps_and_last(membership.read_ranges)
        for first, gap_len in gaps:
            first_local = first
            while gap_len > 0:
                if gap_len > 255:
                    n = 255
                else:
                    n = gap_len
                gap_len -= n
                try:
                    mapping = self.request(
                        requests.ReqLocalToGlobal(membership.conference, first_local, n))
                    unread.extend([e[1] for e in mapping.list if e[1] != 0])
                    first_local = mapping.range_end
                    more_to_fetch = mapping.later_texts_exists
                except NoSuchLocalText:
                    more_to_fetch = 0
        
        # If there are more than 255 after the last read range, we
        # need to continue mapping (the last, if any, previous call
        # will have set later_texts_exists to 1 if so).
        first_local = last
        while more_to_fetch:
            try:
                mapping = self.request(
                    requests.ReqLocalToGlobal(membership.conference, first_local, 255))
                unread.extend([e[1] for e in mapping.list if e[1] != 0])
                first_local = mapping.range_end
                more_to_fetch = mapping.later_texts_exists
            except NoSuchLocalText:
                # No unread texts
                more_to_fetch = 0
        
        # Remove text that don't exist anymore (text_no == 0)
        return [ text_no for text_no in unread if text_no != 0]

    def mark_text(self, text_no, mark_type):
        self.request(requests.ReqMarkText(text_no, mark_type))
        # textstat.misc_info.no_of_marks is now invalid
        self.textstats.invalidate(text_no)

    def unmark_text(self, text_no):
        self.request(requests.ReqUnmarkText(text_no))
        # textstat.misc_info.no_of_marks is now invalid
        self.textstats.invalidate(text_no)


class CachingPersonClient(CachingClient):
    def __init__(self, connection):
        CachingClient.__init__(self, connection)

#    def connect(self, host, port = 4894, user = "", localbind=None):
#        CachingClient.connect(self, host, port, user, localbind)

        # Current person number
        self._pers_no = 0
        
        # Current conference (change-conference)
        self._current_conference_no = 0
        
        # Caches
        self._memberships = Cache(self._fetch_membership, "Membership")
        
        # Specific membership cache where the keys are the positions
        # in the membership list for the membership, and the values
        # are the memberships. There is a risk with having this cache
        # - you can modify positions and we currently have no way of
        # detecting that (no async messages).
        self._memberships_by_position = dict()

        # Setup up async handlers for invalidating cache entries. Skip
        # sending accept-async until the last call.
        self._add_async_handler(AsyncMessages.LEAVE_CONF, self._cpah_leave_conf)
        self._add_async_handler(AsyncMessages.NEW_MEMBERSHIP, self._cpah_new_membership)
        self.request(requests.ReqAcceptAsync(list(self._async_handlers.keys())))

    def login(self, pers_no, password):
        self.request(requests.ReqLogin(pers_no, password, invisible=0))
        # We need to know the current person to be able to have and
        # invalidate caches.
        self._pers_no = pers_no

    def logout(self):
        self.request(requests.ReqLogout())
        # Invalidate caches that are/were for the current person
        self._pers_no = 0
        self._memberships_by_position = dict()
        self._memberships.invalidate_all()

    def get_person_no(self):
        return self._pers_no

    def is_logged_in(self):
        return self._pers_no != 0

    def change_conference(self, conf_no):
        # When changing conference, the lyskom server will update
        # last-time-read for the membership of the *previous*
        # conference. This means that we need to keep track of the
        # current conference to be able to invalidate the membership
        # correctly.
        prev_conf_no = self._current_conference_no
        self.request(requests.ReqChangeConference(conf_no))
        self._current_conference_no = conf_no
        if prev_conf_no != 0:
            self._invalidate_membership(prev_conf_no)

    def mark_as_read_local(self, conf_no, local_text_no):
        try:
            self.request(requests.ReqMarkAsRead(conf_no, [local_text_no]))
        except NotMember:
            pass

    def mark_as_unread_local(self, conf_no, local_text_no):
        try:
            self.request(requests.ReqMarkAsUnread(conf_no, local_text_no))
        except NotMember:
            pass

    def _get_cached_memberships_by_position(self, first, no_of_confs):
        # Return a list of the cached memberships if we have all of
        # them, otherwise return None. We only return memberships if
        # we had all of them.
        memberships = []
        for pos in range(first, first + no_of_confs):
            if pos not in self._memberships_by_position:
                return None
            memberships.append(self._memberships_by_position[pos])
        return memberships

    def _update_cached_memberships_by_position(self, memberships):
        for m in memberships:
            self._memberships_by_position[m.position] = m
    
    def _invalidate_membership(self, conf_no):
        self._memberships.invalidate(conf_no)
        # Since we only return anything from memberships_by_position
        # if all memberships are found, it means that we can make
        # partial invalidations.
        found_at = None
        for pos in self._memberships_by_position:
            if self._memberships_by_position[pos].conference == conf_no:
                found_at = pos
                break
        if found_at is not None:
            del self._memberships_by_position[found_at]
    
    def get_memberships(self, pers_no, first, no_of_confs, want_read_ranges=False):
        """Get memberships for a person.
        """
        if want_read_ranges:
            return self.request(
                requests.ReqGetMembership11(pers_no, 0, no_of_confs, 1, 0))
        else:
            if pers_no == self._pers_no:
                # We cache the result for the current person and without
                # read ranges, because that is what we can invalidate
                # correctly.
                memberships = self._get_cached_memberships_by_position(first, no_of_confs)
                if memberships is None:
                    memberships = self.request(
                        requests.ReqGetMembership11(
                            self._pers_no, first, no_of_confs, 0, 0))
                    self._update_cached_memberships_by_position(memberships)
                return memberships
            else:
                return self.request(
                    requests.ReqGetMembership11(pers_no, 0, no_of_confs, 0, 0))

    def get_membership(self, pers_no, conf_no, want_read_ranges=False):
        """Get a membership for a person
        """
        if want_read_ranges:
            return self.request(requests.ReqQueryReadTexts(pers_no, conf_no, 1, 0))
        else:
            if pers_no == self._pers_no:
                # If it's a membership for the current person and
                # without read ranges, use the cache.
                return self._memberships[conf_no]
            else:
                return self.request(requests.ReqQueryReadTexts(pers_no, conf_no, 0, 0))

    def _fetch_membership(self, conf_no):
        """Fetch the membership for a conf the current person. Does not
        include read ranges.
        """
        # We can only cache memberships for the currently logged in
        # person, because we don't receive async leave/join messages
        # for other persons. We also only cache memberships without
        # read ranges, because it is easier to invalidate correctly.
        return self.request(requests.ReqQueryReadTexts11(self._pers_no, conf_no, 0, 0))
    
    # Handlers for asynchronous messages (internal use)
    def _cpah_leave_conf(self, msg):
        # Invalidates cached membership
        self._memberships.invalidate(msg.conf_no)
        # We invalidate the entire memberships_by_position because you
        # can change position of memberships.
        self._memberships_by_position = dict()

    def _cpah_new_membership(self, msg):
        # Invalidates self._memberships_by_position
        self._memberships_by_position = dict()
        # The self.memberships cache can only cache actual
        # memberships, and because we get this async messages, we know
        # the current person was not a member before.

    # Report cache usage
    def report_cache_usage(self):
        CachingClient.report_cache_usage(self)
        self._memberships.report()


# Cache class for use internally by CachingClient
class Cache(object):
    def __init__(self, fetcher, name = "Unknown"):
        self.dict = {}
        self.fetcher = fetcher
        self.cached = 0
        self.uncached = 0
        self.name = name

    def __getitem__(self, no):
        #print('%s[%d]' % (self.name, no))
        stats.set('clients.cache.{}.gets.last'.format(self.name), 1, agg='sum')
        if no in self.dict:
            #print('%s[%d] - cached' % (self.name, no))
            self.cached = self.cached + 1
            stats.set('clients.cache.{}.gets.hits.last'.format(self.name), 1, agg='sum')
            return self.dict[no]
        else:
            #print('%s[%d] - not cached' % (self.name, no))
            self.uncached = self.uncached + 1
            stats.set('clients.cache.{}.gets.misses.last'.format(self.name), 1, agg='sum')
            self[no] = self.fetcher(no)
            return self.dict[no]

    def __setitem__(self, no, val):
        self.dict[no] = val
        stats.set('clients.cache.{}.sets.last'.format(self.name), 1, agg='sum')

    def invalidate(self, no):
        if no in self.dict:
            del self.dict[no]
            stats.set('clients.cache.{}.invalidations.last'.format(self.name), 1, agg='sum')

    def invalidate_all(self):
        self.dict = dict()
        stats.set('clients.cache.{}.invalidate-alls.last'.format(self.name), 1, agg='sum')

    def report(self):
        print(("Cache %s: %d cached, %d uncached" % (self.name,
                                                     self.cached,
                                                     self.uncached)))



# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.

from .protocol import MAX_TEXT_SIZE

from .datatypes import (
    MIR_TO,
    AnyConfType,
    ArrayAuxItemInput,
    ArrayDynamicSessionInfo,
    ArrayInt32,
    ArrayLocalTextNo,
    ArrayMark,
    ArrayMember,
    ArrayMembership11,
    ArrayMembership10,
    ArrayReadRange,
    ArrayStats,
    ArrayConfNo,
    ArrayConfZInfo,
    AuxNo,
    Bool,
    ConfNo,
    Conference,
    CookedMiscInfo,
    EmptyResponse,
    GarbNice,
    Info,
    InfoType,
    InfoOld,
    Int8,
    Int16,
    Int32,
    LocalTextNo,
    Membership10,
    Membership11,
    MembershipType,
    PersNo,
    Person,
    PersonalFlags,
    PrivBits,
    SchedulingInfo,
    SessionNo,
    StaticServerInfo,
    StaticSessionInfo,
    StatsDescription,
    String,
    TextList,
    TextMapping,
    TextNo,
    TextStat,
    Time,
    UConference,
    VersionInfo)


class Requests(object):
    """Enum like object which give names to the request call numbers
    in the protocol.
    """
    LOGOUT = 1 # (1) Recommended
    CHANGE_CONFERENCE = 2 # (1) Recommended
    CHANGE_NAME = 3 # (1) Recommended
    CHANGE_WHAT_I_AM_DOING = 4 # (1) Recommended
    SET_PRIV_BITS = 7 # (1) Recommended
    SET_PASSWD = 8 # (1) Recommended
    DELETE_CONF = 11 # (1) Recommended
    SUB_MEMBER = 15 # (1) Recommended
    SET_PRESENTATION = 16 # (1) Recommended
    SET_ETC_MOTD = 17 # (1) Recommended
    SET_SUPERVISOR = 18 # (1) Recommended
    SET_PERMITTED_SUBMITTERS = 19 # (1) Recommended
    SET_SUPER_CONF = 20 # (1) Recommended
    SET_CONF_TYPE = 21 # (1) Recommended
    SET_GARB_NICE = 22 # (1) Recommended
    GET_MARKS = 23 # (1) Recommended
    GET_TEXT = 25 # (1) Recommended
    MARK_AS_READ = 27 # (1) Recommended
    DELETE_TEXT = 29 # (1) Recommended
    ADD_RECIPIENT = 30 # (1) Recommended
    SUB_RECIPIENT = 31 # (1) Recommended
    ADD_COMMENT = 32 # (1) Recommended
    SUB_COMMENT = 33 # (1) Recommended
    GET_MAP = 34 # (1) Obsolete (10) Use LOCAL_TO_GLOBAL (103)
    GET_TIME = 35 # (1) Recommended
    ADD_FOOTNOTE = 37 # (1) Recommended
    SUB_FOOTNOTE = 38 # (1) Recommended
    SET_UNREAD = 40 # (1) Recommended
    SET_MOTD_OF_LYSKOM = 41 # (1) Recommended
    ENABLE = 42 # (1) Recommended
    SYNC_KOM = 43 # (1) Recommended
    SHUTDOWN_KOM = 44 # (1) Recommended
    GET_PERSON_STAT = 49 # (1) Recommended
    GET_UNREAD_CONFS = 52 # (1) Recommended
    SEND_MESSAGE = 53 # (1) Recommended
    DISCONNECT = 55 # (1) Recommended
    WHO_AM_I = 56 # (1) Recommended
    SET_USER_AREA = 57 # (2) Recommended
    GET_LAST_TEXT = 58 # (3) Recommended
    FIND_NEXT_TEXT_NO = 60 # (3) Recommended
    FIND_PREVIOUS_TEXT_NO = 61 # (3) Recommended
    LOGIN = 62 # (4) Recommended
    SET_CLIENT_VERSION = 69 # (6) Recommended
    GET_CLIENT_NAME = 70 # (6) Recommended
    GET_CLIENT_VERSION = 71 # (6) Recommended
    MARK_TEXT = 72 # (4) Recommended
    UNMARK_TEXT = 73 # (6) Recommended
    RE_Z_LOOKUP = 74 # (7) Recommended
    GET_VERSION_INFO = 75 # (7) Recommended
    LOOKUP_Z_NAME = 76 # (7) Recommended
    SET_LAST_READ = 77 # (8) Recommended
    GET_UCONF_STAT = 78 # (8) Recommended
    SET_INFO = 79 # (9) Recommended
    ACCEPT_ASYNC = 80 # (9) Recommended
    QUERY_ASYNC = 81 # (9) Recommended
    USER_ACTIVE = 82 # (9) Recommended
    WHO_IS_ON_DYNAMIC = 83 # (9) Recommended
    GET_STATIC_SESSION_INFO = 84 # (9) Recommended
    GET_COLLATE_TABLE = 85 # (10) Recommended
    CREATE_TEXT = 86 # (10) Recommended
    CREATE_ANONYMOUS_TEXT = 87 # (10) Recommended
    CREATE_CONF = 88 # (10) Recommended
    CREATE_PERSON = 89 # (10) Recommended
    GET_TEXT_STAT = 90 # (10) Recommended
    GET_CONF_STAT = 91 # (10) Recommended
    MODIFY_TEXT_INFO = 92 # (10) Recommended
    MODIFY_CONF_INFO = 93 # (10) Recommended
    GET_INFO = 94 # (10) Recommended
    MODIFY_SYSTEM_INFO = 95 # (10) Recommended
    QUERY_PREDEFINED_AUX_ITEMS = 96 # (10) Recommended
    SET_EXPIRE = 97 # (10) Experimental
    QUERY_READ_TEXTS_10 = 98 # (10) Obsolete (11) Use QUERY_READ_TEXTS (107)
    GET_MEMBERSHIP_10 = 99 # (10) Obsolete (11) Use GET_MEMBERSHIP (108)
    ADD_MEMBER = 100 # (10) Recommended
    GET_MEMBERS = 101 # (10) Recommended
    SET_MEMBERSHIP_TYPE = 102 # (10) Recommended
    LOCAL_TO_GLOBAL = 103 # (10) Recommended
    MAP_CREATED_TEXTS = 104 # (10) Recommended
    SET_KEEP_COMMENTED = 105 # (11) Recommended (10) Experimental
    SET_PERS_FLAGS = 106 # (10) Recommended
    QUERY_READ_TEXTS = 107 # (11) Recommended
    GET_MEMBERSHIP = 108 # (11) Recommended
    MARK_AS_UNREAD = 109 # (11) Recommended
    SET_READ_RANGES = 110 # (11) Recommended
    GET_STATS_DESCRIPTION = 111 # (11) Recommended
    GET_STATS = 112 # (11) Recommended
    GET_BOOTTIME_INFO = 113 # (11) Recommended
    FIRST_UNUSED_CONF_NO = 114 # (11) Recommended
    FIRST_UNUSED_TEXT_NO = 115 # (11) Recommended
    FIND_NEXT_CONF_NO = 116 # (11) Recommended
    FIND_PREVIOUS_CONF_NO = 117 # (11) Recommended
    GET_SCHEDULING = 118 # (11) Experimental
    SET_SCHEDULING = 119 # (11) Experimental
    SET_CONNECTION_TIME_FORMAT = 120 # (11) Recommended
    LOCAL_TO_GLOBAL_REVERSE = 121 # (11) Recommended
    MAP_CREATED_TEXTS_REVERSE = 122 # (11) Recommended




#
# Classes for requests to the server are all subclasses of Request.
#

class Argument(object):
    def __init__(self, name, data_type, default=None):
        self.name = name
        self.data_type = data_type
        self.default = default

    def __repr__(self):
        return "Argument({!r}, {!r}, default={!r})".format(
            self.name, self.data_type, self.default)


class Request(object):
    CALL_NO = None # Override - Integer protocol request call number.
    ARGS = None # Override - List of Argument(s).

    def __init__(self, *args, **kwargs):
        """
        @param *args Arguments supplied in the same order as ARGS.

        @param **kwargs Key-word arguments with the names specified in ARGS.
        """
        if self.CALL_NO is None:
            raise TypeError("Must have CALL_NO")

        if self.ARGS is None:
            raise TypeError("Must have ARGS")

        args_count = len(self.ARGS)
        given_args = len(args)
        given_kwargs = len(kwargs)
        given_total = given_args + given_kwargs
        if given_total > args_count:
            raise TypeError("Takes at most {:d} arguments ({:d} given)".format(
                    args_count, given_total))

        self.args = []
        self._serialized_args = []
        for i, arg_def in enumerate(self.ARGS):
            if i < given_args:
                val = args[i]
                if arg_def.name in kwargs:
                    raise TypeError("Argument {} specified both "
                                    "as positional and keyword argument".format(
                            arg_def.name))
            else:
                if arg_def.name in kwargs:
                    val = kwargs[arg_def.name]
                else:
                    if arg_def.default is None:
                        raise TypeError("Argument {} is missing".format(arg_def.name))
                    else:
                        val = arg_def.default

            assert not hasattr(self, arg_def.name), "Invalid argument name"

            # FIXME. This calls the constructor of the datatype with
            # the value as argument.  this might make sense for
            # converting ints and strings to Int32 and String (and
            # similar), but not for complex types, which are expected
            # to be passed as an argument already. In those cases it
            # will just call the constructor of the datatype class
            # with an instance of that type.
            #
            # TODO: call a classmethod on the types instead of the
            # constructor.
            arg = arg_def.data_type(val)

            setattr(self, arg_def.name, arg)
            self.args.append(arg)
            self._serialized_args.append(arg.to_string())

    # TODO: Rename to "to_bytes"
    def to_string(self):
        """Returns the full serialized request, including CALL_NO and
        end of line. To bytes.
        """
        return b' '.join([b"%d" % (self.CALL_NO, )] + self._serialized_args) + b"\n"


    def __repr__(self):
        cls_name = type(self).__name__
        arg_values = []
        for i, arg_def in enumerate(self.ARGS):
            arg_values.append("{}={!r}".format(arg_def.name, self.args[i]))
        return "{cls_name}({args})".format(
            cls_name=cls_name,
            args=", ".join(arg_values))


# login-old [0] (1) Obsolete (4) Use login (62)

# logout [1] (1) Recommended
class ReqLogout(Request):
    CALL_NO = Requests.LOGOUT
    ARGS = []

# change-conference [2] (1) Recommended
class ReqChangeConference(Request):
    CALL_NO = Requests.CHANGE_CONFERENCE
    ARGS = [ Argument('conf_no', Int16) ]

# change-name [3] (1) Recommended
class ReqChangeName(Request):
    CALL_NO = Requests.CHANGE_NAME
    ARGS = [ Argument('conf_no', Int16),
             Argument('new_name', String) ]

# change-what-i-am-doing [4] (1) Recommended
class ReqChangeWhatIAmDoing(Request):
    CALL_NO = Requests.CHANGE_WHAT_I_AM_DOING
    ARGS = [ Argument('what', String) ]

# create-person-old [5] (1) Obsolete (10) Use create-person (89)
# get-person-stat-old [6] (1) Obsolete (1) Use get-person-stat (49)

# set-priv-bits [7] (1) Recommended
class ReqSetPrivBits(Request):
    CALL_NO = Requests.SET_PRIV_BITS
    ARGS = [ Argument('person', PersNo),
             Argument('privileges', PrivBits) ]

# set-passwd [8] (1) Recommended
class ReqSetPasswd(Request):
    CALL_NO = Requests.SET_PASSWD
    ARGS = [ Argument('person', PersNo),
             Argument('old_pwd', String),
             Argument('new_pwd', String) ]

# query-read-texts-old [9] (1) Obsolete (10) Use query-read-texts (98)
# create-conf-old [10] (1) Obsolete (10) Use create-conf (88)

# delete-conf [11] (1) Recommended
class ReqDeleteConf(Request):
    CALL_NO = Requests.DELETE_CONF
    ARGS = [ Argument('conf', ConfNo) ]

# lookup-name [12] (1) Obsolete (7) Use lookup-z-name (76)
# get-conf-stat-older [13] (1) Obsolete (10) Use get-conf-stat (91)
# add-member-old [14] (1) Obsolete (10) Use add-member (100)

# sub-member [15] (1) Recommended
class ReqSubMember(Request):
    CALL_NO = Requests.SUB_MEMBER
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('pers_no', PersNo) ]

# set-presentation [16] (1) Recommended
class ReqSetPresentation(Request):
    CALL_NO = Requests.SET_PRESENTATION
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('text_no', TextNo) ]

# set-etc-motd [17] (1) Recommended
class ReqSetEtcMoTD(Request):
    CALL_NO = Requests.SET_ETC_MOTD
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('text_no', TextNo) ]

# set-supervisor [18] (1) Recommended
class ReqSetSupervisor(Request):
    CALL_NO = Requests.SET_SUPERVISOR
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('admin', ConfNo) ]

# set-permitted-submitters [19] (1) Recommended
class ReqSetPermittedSubmitters(Request):
    CALL_NO = Requests.SET_PERMITTED_SUBMITTERS
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('perm_sub', ConfNo) ]

# set-super-conf [20] (1) Recommended
class ReqSetSuperConf(Request):
    CALL_NO = Requests.SET_SUPER_CONF
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('super_conf', ConfNo) ]

# set-conf-type [21] (1) Recommended
class ReqSetConfType(Request):
    CALL_NO = Requests.SET_CONF_TYPE
    ARGS = [ Argument('conf_no', ConfNo),
             # We use AnyConfType, which means always sending as
             # ExtendedConfType, even if a ConfType is supplied.
             Argument('conf_type', AnyConfType) ]

# set-garb-nice [22] (1) Recommended
class ReqSetGarbNice(Request):
    CALL_NO = Requests.SET_GARB_NICE
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('nice', GarbNice) ]

# get-marks [23] (1) Recommended
class ReqGetMarks(Request):
    CALL_NO = Requests.GET_MARKS
    ARGS = []

# mark-text-old [24] (1) Obsolete (4) Use mark-text/unmark-text (72/73)

# get-text [25] (1) Recommended
class ReqGetText(Request):
    CALL_NO = Requests.GET_TEXT
    ARGS = [ Argument('text_no', TextNo),
             Argument('start_char', Int32, default=0),
             Argument('end_char', Int32, default=MAX_TEXT_SIZE) ]
    
# get-text-stat-old [26] (1) Obsolete (10) Use get-text-stat (90)

# mark-as-read [27] (1) Recommended
class ReqMarkAsRead(Request):
    CALL_NO = Requests.MARK_AS_READ
    ARGS = [ Argument('conf_no', ConfNo), 
             Argument('texts', ArrayLocalTextNo) ]
                      
# create-text-old [28] (1) Obsolete (10) Use create-text (86)

# delete-text [29] (1) Recommended
class ReqDeleteText(Request):
    CALL_NO = Requests.DELETE_TEXT
    ARGS = [ Argument('text_no', TextNo) ]

# add-recipient [30] (1) Recommended
class ReqAddRecipient(Request):
    CALL_NO = Requests.ADD_RECIPIENT
    ARGS = [ Argument('text_no', TextNo),
             Argument('conf_no', ConfNo),
             Argument('recpt_type', InfoType, default=MIR_TO) ]

# sub-recipient [31] (1) Recommended
class ReqSubRecipient(Request):
    CALL_NO = Requests.SUB_RECIPIENT
    ARGS = [ Argument('text_no', TextNo),
             Argument('conf_no', ConfNo) ]

# add-comment [32] (1) Recommended
class ReqAddComment(Request):
    CALL_NO = Requests.ADD_COMMENT
    ARGS = [ Argument('text_no', TextNo),
             Argument('comment_to', TextNo) ]

# sub-comment [33] (1) Recommended
class ReqSubComment(Request):
    CALL_NO = Requests.SUB_COMMENT
    ARGS = [ Argument('text_no', TextNo),
             Argument('comment_to', TextNo) ]

# get-map [34] (1) Obsolete (10) Use local-to-global (103)

# get-time [35] (1) Recommended
class ReqGetTime(Request):
    CALL_NO = Requests.GET_TIME
    ARGS = []

# get-info-old [36] (1) Obsolete (10) Use get-info (94)

# add-footnote [37] (1) Recommended
class ReqAddFootnote(Request):
    CALL_NO = Requests.ADD_FOOTNOTE
    ARGS = [ Argument('text_no', TextNo),
             Argument('footnote_to', TextNo) ]

# sub-footnote [38] (1) Recommended
class ReqSubFootnote(Request):
    CALL_NO = Requests.SUB_FOOTNOTE
    ARGS = [ Argument('text_no', TextNo),
             Argument('footnote_to', TextNo) ]

# who-is-on-old [39] (1) Obsolete (9) Use get-static-session-info (84) and
#                                         who-is-on-dynamic (83)

# set-unread [40] (1) Recommended
class ReqSetUnread(Request):
    CALL_NO = Requests.SET_UNREAD
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('no_of_unread', Int32) ]

# set-motd-of-lyskom [41] (1) Recommended
class ReqSetMoTDOfLysKOM(Request):
    CALL_NO = Requests.SET_MOTD_OF_LYSKOM
    ARGS = [ Argument('text_no', TextNo) ]

# enable [42] (1) Recommended
class ReqEnable(Request):
    CALL_NO = Requests.ENABLE
    ARGS = [ Argument('level', Int8) ]

# sync-kom [43] (1) Recommended
class ReqSyncKOM(Request):
    CALL_NO = Requests.SYNC_KOM
    ARGS = []

# shutdown-kom [44] (1) Recommended
class ReqShutdownKOM(Request):
    CALL_NO = Requests.SHUTDOWN_KOM
    ARGS = [ Argument('exit_val', Int8) ]

# broadcast [45] (1) Obsolete (1) Use send-message (53)
# get-membership-old [46] (1) Obsolete (10) Use get-membership (99)
# get-created-texts [47] (1) Obsolete (10) Use map-created-texts (104)
# get-members-old [48] (1) Obsolete (10) Use get-members (101)

# get-person-stat [49] (1) Recommended
class ReqGetPersonStat(Request):
    CALL_NO = Requests.GET_PERSON_STAT
    ARGS = [ Argument('pers_no', PersNo) ]

# get-conf-stat-old [50] (1) Obsolete (10) Use get-conf-stat (91)

# who-is-on [51] (1) Obsolete (9)  Use who-is-on-dynamic (83) and
#                                      get-static-session-info (84)

# get-unread-confs [52] (1) Recommended
class ReqGetUnreadConfs(Request):
    CALL_NO = Requests.GET_UNREAD_CONFS
    ARGS = [ Argument('pers_no', PersNo) ] 

# send-message [53] (1) Recommended
class ReqSendMessage(Request):
    CALL_NO = Requests.SEND_MESSAGE
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('message', String) ]

# get-session-info [54] (1) Obsolete (9) Use who-is-on-dynamic (83)

# disconnect [55] (1) Recommended
class ReqDisconnect(Request):
    CALL_NO = Requests.DISCONNECT
    ARGS = [ Argument('session_no', SessionNo) ]

# who-am-i [56] (1) Recommended
class ReqWhoAmI(Request):
    CALL_NO = Requests.WHO_AM_I
    ARGS = []

# set-user-area [57] (2) Recommended
class ReqSetUserArea(Request):
    CALL_NO = Requests.SET_USER_AREA
    ARGS = [ Argument('pers_no', PersNo),
             Argument('user_area', TextNo) ]

# get-last-text [58] (3) Recommended
class ReqGetLastText(Request):
    CALL_NO = Requests.GET_LAST_TEXT
    ARGS = [ Argument('before', Time) ]

# create-anonymous-text-old [59] (3) Obsolete (10)
#                                    Use create-anonymous-text (87)

# find-next-text-no [60] (3) Recommended
class ReqFindNextTextNo(Request):
    CALL_NO = Requests.FIND_NEXT_TEXT_NO
    ARGS = [ Argument('start', TextNo) ]

# find-previous-text-no [61] (3) Recommended
class ReqFindPreviousTextNo(Request):
    CALL_NO = Requests.FIND_PREVIOUS_TEXT_NO
    ARGS = [ Argument('start', TextNo) ]

# login [62] (4) Recommended
class ReqLogin(Request):
    CALL_NO = Requests.LOGIN
    ARGS = [ Argument('person', PersNo),
             Argument('password', String),
             Argument('invisible', Bool, default=1) ]

# who-is-on-ident [63] (4) Obsolete (9) Use who-is-on-dynamic (83) and
#                                           get-static-session-info (84)
# get-session-info-ident [64] (4) Obsolete (9) Use who-is-on-dynamic (83) and
#                                              get-static-session-info (84)
# re-lookup-person [65] (5) Obsolete (7) Use re-z-lookup (74)
# re-lookup-conf [66] (5) Obsolete (7) Use re-z-lookup (74)
# lookup-person [67] (6) Obsolete (7) Use lookup-z-name (76)
# lookup-conf [68] (6) Obsolete (7) Use lookup-z-name (76)

# set-client-version [69] (6) Recommended
class ReqSetClientVersion(Request):
    CALL_NO = Requests.SET_CLIENT_VERSION
    ARGS = [ Argument('client_name', String),
             Argument('client_version', String) ]

# get-client-name [70] (6) Recommended
class ReqGetClientName(Request):
    CALL_NO = Requests.GET_CLIENT_NAME
    ARGS = [ Argument('session_no', SessionNo) ]

# get-client-version [71] (6) Recommended
class ReqGetClientVersion(Request):
    CALL_NO = Requests.GET_CLIENT_VERSION
    ARGS = [ Argument('session_no', SessionNo) ]

# mark-text [72] (4) Recommended
class ReqMarkText(Request):
    CALL_NO = Requests.MARK_TEXT
    ARGS = [ Argument('text_no', TextNo),
             Argument('mark_type', Int8) ]

# unmark-text [73] (6) Recommended
class ReqUnmarkText(Request):
    CALL_NO = Requests.UNMARK_TEXT
    ARGS = [ Argument('text_no', TextNo) ]

# re-z-lookup [74] (7) Recommended
class ReqReZLookup(Request):
    CALL_NO = Requests.RE_Z_LOOKUP
    ARGS = [ Argument('regexp', String),
             Argument('want_persons', Bool),
             Argument('want_confs', Bool) ]

# get-version-info [75] (7) Recommended
class ReqGetVersionInfo(Request):
    CALL_NO = Requests.GET_VERSION_INFO
    ARGS = []

# lookup-z-name [76] (7) Recommended
class ReqLookupZName(Request):
    CALL_NO = Requests.LOOKUP_Z_NAME
    ARGS = [ Argument('name', String),
             Argument('want_pers', Bool),
             Argument('want_confs', Bool) ]

# set-last-read [77] (8) Recommended
class ReqSetLastRead(Request):
    CALL_NO = Requests.SET_LAST_READ
    ARGS = [ Argument('conference', ConfNo),
             Argument('last_read', LocalTextNo) ]

# get-uconf-stat [78] (8) Recommended
class ReqGetUconfStat(Request):
    CALL_NO = Requests.GET_UCONF_STAT
    ARGS = [ Argument('conf_no', ConfNo) ]

# set-info [79] (9) Recommended
class ReqSetInfo(Request):
    CALL_NO = Requests.SET_INFO
    ARGS = [ Argument('info', InfoOld) ]

# accept-async [80] (9) Recommended
class ReqAcceptAsync(Request):
    CALL_NO = Requests.ACCEPT_ASYNC
    ARGS = [ Argument('request_list', ArrayInt32) ]

# query-async [81] (9) Recommended
class ReqQueryAsync(Request):
    CALL_NO = Requests.QUERY_ASYNC
    ARGS = []

# user-active [82] (9) Recommended
class ReqUserActive(Request):
    CALL_NO = Requests.USER_ACTIVE
    ARGS = []

# who-is-on-dynamic [83] (9) Recommended
class ReqWhoIsOnDynamic(Request):
    CALL_NO = Requests.WHO_IS_ON_DYNAMIC
    ARGS = [ Argument('want_visible', Bool),
             Argument('want_invisible', Bool),
             Argument('active_last', Int32) ]

# get-static-session-info [84] (9) Recommended
class ReqGetStaticSessionInfo(Request):
    CALL_NO = Requests.GET_STATIC_SESSION_INFO
    ARGS = [ Argument('session_no', SessionNo) ]

# get-collate-table [85] (10) Recommended
class ReqGetCollateTable(Request):
    CALL_NO = Requests.GET_COLLATE_TABLE
    ARGS = []

# create-text [86] (10) Recommended
class ReqCreateText(Request):
    CALL_NO = Requests.CREATE_TEXT
    ARGS = [ Argument('text', String),
             Argument('misc_info', CookedMiscInfo),
             Argument('aux_items', ArrayAuxItemInput) ]

# create-anonymous-text [87] (10) Recommended
class ReqCreateAnonymousText(Request):
    CALL_NO = Requests.CREATE_ANONYMOUS_TEXT
    ARGS = [ Argument('text', String),
             Argument('misc_info', CookedMiscInfo),
             Argument('aux_items', ArrayAuxItemInput) ]

# create-conf [88] (10) Recommended
class ReqCreateConf(Request):
    CALL_NO = Requests.CREATE_CONF
    ARGS = [ Argument('name', String),
             Argument('type', AnyConfType),
             Argument('aux_items', ArrayAuxItemInput) ]

# create-person [89] (10) Recommended
class ReqCreatePerson(Request):
    CALL_NO = Requests.CREATE_PERSON
    ARGS = [ Argument('name', String),
             Argument('passwd', String),
             Argument('flags', PersonalFlags),
             Argument('aux_items', ArrayAuxItemInput) ]

# get-text-stat [90] (10) Recommended
class ReqGetTextStat(Request):
    CALL_NO = Requests.GET_TEXT_STAT
    ARGS = [ Argument('text_no', TextNo) ]

# get-conf-stat [91] (10) Recommended
class ReqGetConfStat(Request):
    CALL_NO = Requests.GET_CONF_STAT
    ARGS = [ Argument('conf_no', ConfNo) ]

# modify-text-info [92] (10) Recommended
class ReqModifyTextInfo(Request):
    CALL_NO = Requests.MODIFY_TEXT_INFO
    ARGS = [ Argument('text', TextNo),
             Argument('delete', AuxNo),
             Argument('add', ArrayAuxItemInput) ]

# modify-conf-info [93] (10) Recommended
class ReqModifyConfInfo(Request):
    CALL_NO = Requests.MODIFY_CONF_INFO
    ARGS = [ Argument('conf', ConfNo),
             Argument('delete', AuxNo),
             Argument('add', ArrayAuxItemInput) ]

# get-info [94] (10) Recommended
class ReqGetInfo(Request):
    CALL_NO = Requests.GET_INFO
    ARGS = []

# modify-system-info [95] (10) Recommended
class ReqModifySystemInfo(Request):
    CALL_NO = Requests.MODIFY_SYSTEM_INFO
    ARGS = [ Argument('items_to_delete', AuxNo),
             Argument('items_to_add', ArrayAuxItemInput) ]

# query-predefined-aux-items [96] (10) Recommended
class ReqQueryPredefinedAuxItems(Request):
    CALL_NO = Requests.QUERY_PREDEFINED_AUX_ITEMS
    ARGS = []

# set-expire [97] (10) Experimental
class ReqSetExpire(Request):
    CALL_NO = Requests.SET_EXPIRE
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('expire', GarbNice) ]

# query-read-texts-10 [98] (10) Obsolete (11) Use query-read-texts (107)
class ReqQueryReadTexts10(Request):
    CALL_NO = Requests.QUERY_READ_TEXTS_10
    ARGS = [ Argument('person', PersNo),
             Argument('conference', ConfNo) ]

# get-membership-10 [99] (10) Obsolete (11) Use get-membership (108)

class ReqGetMembership10(Request):
    CALL_NO = Requests.GET_MEMBERSHIP_10
    ARGS = [ Argument('person', PersNo),
             Argument('first', Int16),
             Argument('no_of_confs', Int16),
             Argument('want_read_texts', Bool) ]

# add-member [100] (10) Recommended
class ReqAddMember(Request):
    CALL_NO = Requests.ADD_MEMBER
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('pers_no', PersNo),
             Argument('priority', Int8),
             Argument('where', Int16),
             Argument('type', MembershipType) ]

# get-members [101] (10) Recommended
class ReqGetMembers(Request):
    CALL_NO = Requests.GET_MEMBERS
    ARGS = [ Argument('conf', ConfNo),
             Argument('first', Int16),
             Argument('no_of_members', Int16) ]

# set-membership-type [102] (10) Recommended
class ReqSetMembershipType(Request):
    CALL_NO = Requests.SET_MEMBERSHIP_TYPE
    ARGS = [ Argument('pers', PersNo),
             Argument('conf', ConfNo),
             Argument('type', MembershipType) ]

# local-to-global [103] (10) Recommended
class ReqLocalToGlobal(Request):
    CALL_NO = Requests.LOCAL_TO_GLOBAL
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('first_local_no', LocalTextNo),
             Argument('no_of_existing_texts', Int32) ]

# map-created-texts [104] (10) Recommended
class ReqMapCreatedTexts(Request):
    CALL_NO = Requests.MAP_CREATED_TEXTS
    ARGS = [ Argument('author', PersNo),
             Argument('first_local_no', LocalTextNo),
             Argument('no_of_existing_texts', Int32) ]

# set-keep-commented [105] (11) Recommended (10) Experimental
class ReqSetKeepCommented(Request):
    CALL_NO = Requests.SET_KEEP_COMMENTED
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('keep_commented', GarbNice) ]

# set-pers-flags [106] (10) Recommended

class ReqSetPersFlags(Request):
    CALL_NO = Requests.SET_PERS_FLAGS
    ARGS = [ Argument('pers_no', PersNo),
             Argument('flags', PersonalFlags) ]

### --- New in protocol version 11 ---

# query-read-texts [107] (11) Recommended
class ReqQueryReadTexts11(Request):
    CALL_NO = Requests.QUERY_READ_TEXTS
    ARGS = [ Argument('person', PersNo),
             Argument('conference', ConfNo),
             Argument('want_read_ranges', Bool),
             Argument('max_ranges', Int32) ]

ReqQueryReadTexts = ReqQueryReadTexts11

# get-membership [108] (11) Recommended
class ReqGetMembership11(Request):
    CALL_NO = Requests.GET_MEMBERSHIP
    ARGS = [ Argument('person', PersNo),
             Argument('first', Int16),
             Argument('no_of_confs', Int16),
             Argument('want_read_ranges', Bool),
             Argument('max_ranges', Int32) ]

ReqGetMembership = ReqGetMembership11

# mark-as-unread [109] (11) Recommended
class ReqMarkAsUnread(Request):
    CALL_NO = Requests.MARK_AS_UNREAD
    ARGS = [ Argument('conference', ConfNo),
             Argument('text', LocalTextNo) ]

# set-read-ranges [110] (11) Recommended
class ReqSetReadRanges(Request):
    CALL_NO = Requests.SET_READ_RANGES
    ARGS = [ Argument('conference', ConfNo),
             Argument('read_ranges', ArrayReadRange) ]

# get-stats-description [111] (11) Recommended
class ReqGetStatsDescription(Request):
    CALL_NO = Requests.GET_STATS_DESCRIPTION
    ARGS = []

# get-stats [112] (11) Recommended
class ReqGetStats(Request):
    CALL_NO = Requests.GET_STATS
    ARGS = [ Argument('what', String) ]

# get-boottime-info [113] (11) Recommended
class ReqGetBoottimeInfo(Request):
    CALL_NO = Requests.GET_BOOTTIME_INFO
    ARGS = []

# first-unused-conf-no [114] (11) Recommended
class ReqFirstUnusedConfNo(Request):
    CALL_NO = Requests.FIRST_UNUSED_CONF_NO
    ARGS = []

# first-unused-text-no [115] (11) Recommended
class ReqFirstUnusedTextNo(Request):
    CALL_NO = Requests.FIRST_UNUSED_TEXT_NO
    ARGS = []

# find-next-conf-no [116] (11) Recommended
class ReqFindNextConfNo(Request):
    CALL_NO = Requests.FIND_NEXT_CONF_NO
    ARGS = [ Argument('start', ConfNo) ]

# find-previous-conf-no [117] (11) Recommended
class ReqFindPreviousConfNo(Request):
    CALL_NO = Requests.FIND_PREVIOUS_CONF_NO
    ARGS = [ Argument('start', ConfNo) ]

# get-scheduling [118] (11) Experimental
class ReqGetScheduling(Request):
    CALL_NO = Requests.GET_SCHEDULING
    ARGS = [ Argument('session_no', SessionNo) ]

# set-scheduling [119] (11) Experimental
class ReqSetScheduling(Request):
    CALL_NO = Requests.SET_SCHEDULING
    ARGS = [ Argument('session_no', SessionNo),
             Argument('priority', Int16),
             Argument('weight', Int16) ]

# set-connection-time-format [120] (11) Recommended
class ReqSetConnectionTimeFormat(Request):
    CALL_NO = Requests.SET_CONNECTION_TIME_FORMAT
    ARGS = [ Argument('use_utc', Bool) ]

# local-to-global-reverse [121] (11) Recommended
class ReqLocalToGlobalReverse(Request):
    CALL_NO = Requests.LOCAL_TO_GLOBAL_REVERSE
    ARGS = [ Argument('conf_no', ConfNo),
             Argument('local_no_ceiling', LocalTextNo),
             Argument('no_of_existing_texts', Int32) ]

# map-created-texts-reverse [122] (11) Recommended
class ReqMapCreatedTextsReverse(Request):
    CALL_NO = Requests.MAP_CREATED_TEXTS_REVERSE
    ARGS = [ Argument('author', PersNo),
             Argument('local_no_ceiling', LocalTextNo),
             Argument('no_of_existing_texts', Int32) ]



# Mapping from request type to response type.
response_dict = {
    Requests.ACCEPT_ASYNC: EmptyResponse,
    Requests.ADD_COMMENT: EmptyResponse,
    Requests.ADD_FOOTNOTE: EmptyResponse,
    Requests.ADD_MEMBER: EmptyResponse,
    Requests.ADD_RECIPIENT: EmptyResponse,
    Requests.CHANGE_CONFERENCE: EmptyResponse,
    Requests.CHANGE_NAME: EmptyResponse,
    Requests.CHANGE_WHAT_I_AM_DOING: EmptyResponse,
    Requests.CREATE_ANONYMOUS_TEXT: TextNo,
    Requests.CREATE_CONF: ConfNo,
    Requests.CREATE_PERSON: PersNo,
    Requests.CREATE_TEXT: TextNo,
    Requests.DELETE_CONF: EmptyResponse,
    Requests.DELETE_TEXT: EmptyResponse,
    Requests.DISCONNECT: EmptyResponse,
    Requests.ENABLE: EmptyResponse,
    Requests.FIND_NEXT_CONF_NO: ConfNo,
    Requests.FIND_NEXT_TEXT_NO: TextNo,
    Requests.FIND_PREVIOUS_CONF_NO: ConfNo,
    Requests.FIND_PREVIOUS_TEXT_NO: TextNo,
    Requests.FIRST_UNUSED_CONF_NO: ConfNo,
    Requests.FIRST_UNUSED_TEXT_NO: TextNo,
    Requests.GET_BOOTTIME_INFO: StaticServerInfo,
    Requests.GET_CLIENT_NAME: String,
    Requests.GET_CLIENT_VERSION: String,
    Requests.GET_COLLATE_TABLE: String,
    Requests.GET_CONF_STAT: Conference,
    Requests.GET_INFO: EmptyResponse,
    Requests.GET_INFO: Info,
    Requests.GET_LAST_TEXT: TextNo,
    Requests.GET_MAP: TextList,
    Requests.GET_MARKS: ArrayMark,
    Requests.GET_MEMBERS: ArrayMember,
    Requests.GET_MEMBERSHIP: ArrayMembership11,
    Requests.GET_MEMBERSHIP_10: ArrayMembership10,
    Requests.GET_PERSON_STAT: Person,
    Requests.GET_SCHEDULING: SchedulingInfo,
    Requests.GET_STATIC_SESSION_INFO: StaticSessionInfo,
    Requests.GET_STATS: ArrayStats,
    Requests.GET_STATS_DESCRIPTION: StatsDescription,
    Requests.GET_TEXT: String,
    Requests.GET_TEXT_STAT: TextStat,
    Requests.GET_TIME: Time,
    Requests.GET_UCONF_STAT: UConference,
    Requests.GET_UNREAD_CONFS: ArrayConfNo,
    Requests.GET_VERSION_INFO: VersionInfo,
    Requests.LOCAL_TO_GLOBAL: TextMapping,
    Requests.LOCAL_TO_GLOBAL_REVERSE: TextMapping,
    Requests.LOGIN: EmptyResponse,
    Requests.LOGOUT: EmptyResponse,
    Requests.LOOKUP_Z_NAME: ArrayConfZInfo,
    Requests.MAP_CREATED_TEXTS: TextMapping,
    Requests.MAP_CREATED_TEXTS_REVERSE: TextMapping,
    Requests.MARK_AS_READ: EmptyResponse,
    Requests.MARK_AS_UNREAD: EmptyResponse,
    Requests.MARK_TEXT: EmptyResponse,
    Requests.MODIFY_CONF_INFO: EmptyResponse,
    Requests.MODIFY_SYSTEM_INFO: EmptyResponse,
    Requests.MODIFY_TEXT_INFO: EmptyResponse,
    Requests.QUERY_ASYNC: ArrayInt32,
    Requests.QUERY_PREDEFINED_AUX_ITEMS: ArrayInt32,
    Requests.QUERY_READ_TEXTS: Membership11,
    Requests.QUERY_READ_TEXTS_10: Membership10,
    Requests.RE_Z_LOOKUP: ArrayConfZInfo,
    Requests.SEND_MESSAGE: EmptyResponse,
    Requests.SET_CLIENT_VERSION: EmptyResponse,
    Requests.SET_CONF_TYPE: EmptyResponse,
    Requests.SET_CONNECTION_TIME_FORMAT: EmptyResponse,
    Requests.SET_ETC_MOTD: EmptyResponse,
    Requests.SET_EXPIRE: EmptyResponse,
    Requests.SET_GARB_NICE: EmptyResponse,
    Requests.SET_INFO: EmptyResponse,
    Requests.SET_KEEP_COMMENTED: EmptyResponse,
    Requests.SET_LAST_READ: EmptyResponse,
    Requests.SET_MEMBERSHIP_TYPE: EmptyResponse,
    Requests.SET_MOTD_OF_LYSKOM: EmptyResponse,
    Requests.SET_PASSWD: EmptyResponse,
    Requests.SET_PERMITTED_SUBMITTERS: EmptyResponse,
    Requests.SET_PERS_FLAGS: EmptyResponse,
    Requests.SET_PRESENTATION: EmptyResponse,
    Requests.SET_PRIV_BITS: EmptyResponse,
    Requests.SET_READ_RANGES: EmptyResponse,
    Requests.SET_SCHEDULING: EmptyResponse,
    Requests.SET_SUPERVISOR: EmptyResponse,
    Requests.SET_SUPER_CONF: EmptyResponse,
    Requests.SET_UNREAD: EmptyResponse,
    Requests.SET_USER_AREA: EmptyResponse,
    Requests.SHUTDOWN_KOM: EmptyResponse,
    Requests.SUB_COMMENT: EmptyResponse,
    Requests.SUB_FOOTNOTE: EmptyResponse,
    Requests.SUB_MEMBER: EmptyResponse,
    Requests.SUB_RECIPIENT: EmptyResponse,
    Requests.SYNC_KOM: EmptyResponse,
    Requests.UNMARK_TEXT: EmptyResponse,
    Requests.USER_ACTIVE: EmptyResponse,
    Requests.WHO_AM_I: SessionNo,
    Requests.WHO_IS_ON_DYNAMIC: ArrayDynamicSessionInfo,
}



# AUTOGENERATED by make_komauxitems from aux-items.py

AI_CONTENT_TYPE = 1 # (text)
AI_FAST_REPLY = 2 # (text)
AI_CROSS_REFERENCE = 3 # (text, conference, letterbox)
AI_NO_COMMENTS = 4 # (text)
AI_PERSONAL_COMMENT = 5 # (text)
AI_REQUEST_CONFIRMATION = 6 # (text)
AI_READ_CONFIRM = 7 # (text)
AI_REDIRECT = 8 # (conference, letterbox)
AI_X_FACE = 9 # (conference, letterbox, server)
AI_ALTERNATE_NAME = 10 # (text, conference, letterbox)
AI_PGP_SIGNATURE = 11 # (text)
AI_PGP_PUBLIC_KEY = 12 # (letterbox)
AI_E_MAIL_ADDRESS = 13 # (conference, letterbox, server)
AI_FAQ_TEXT = 14 # (conference, letterbox, server)
AI_CREATING_SOFTWARE = 15 # (text)
AI_MX_AUTHOR = 16 # (text)
AI_MX_FROM = 17 # (text)
AI_MX_REPLY_TO = 18 # (text)
AI_MX_TO = 19 # (text)
AI_MX_CC = 20 # (text)
AI_MX_DATE = 21 # (text)
AI_MX_MESSAGE_ID = 22 # (text)
AI_MX_IN_REPLY_TO = 23 # (text)
AI_MX_MISC = 24 # (text)
AI_MX_ALLOW_FILTER = 25 # (conference, letterbox)
AI_MX_REJECT_FORWARD = 26 # (conference, letterbox)
AI_NOTIFY_COMMENTS = 27 # (letterbox)
AI_FAQ_FOR_CONF = 28 # (text)
AI_RECOMMENDED_CONF = 29 # (server)
AI_ALLOWED_CONTENT_TYPE = 30 # (conference, letterbox, server)
AI_CANONICAL_NAME = 31 # (server)
AI_MX_LIST_NAME = 32 # (conference)
AI_SEND_COMMENTS_TO = 33 # (letterbox)
AI_WORLD_READABLE = 34 # (text)
AI_MX_REFUSE_IMPORT = 35 # (conference, letterbox)
AI_ELISP_CLIENT_READ_FAQ = 10000 # (letterbox)
AI_ELISP_CLIENT_REJECTED_RECOMMENDATION = 10001 # (letterbox)
AI_MX_MIME_BELONGS_TO = 10100 # (text)
AI_MX_MIME_PART_IN = 10101 # (text)
AI_MX_MIME_MISC = 10102 # (text)
AI_MX_ENVELOPE_SENDER = 10103 # (text)
AI_MX_MIME_FILE_NAME = 10104 # (text)
AI_MX_ALLOW_ENVELOPE_SENDER_REGEXP = 10105 # (conference, letterbox)
AI_KOMFEEDER_UNIQUE_ID = 10400 # (text)
AI_KOMFEEDER_URL = 10401 # (conference, text)
AI_KOMFEEDER_TITLE = 10402 # (conference)
AI_KOMFEEDER_AUTHOR = 10403 # (text)
AI_KOMFEEDER_DATE = 10404 # (text)

aux_item_number_to_name = {
  AI_CONTENT_TYPE: "content-type",
  AI_FAST_REPLY: "fast-reply",
  AI_CROSS_REFERENCE: "cross-reference",
  AI_NO_COMMENTS: "no-comments",
  AI_PERSONAL_COMMENT: "personal-comment",
  AI_REQUEST_CONFIRMATION: "request-confirmation",
  AI_READ_CONFIRM: "read-confirm",
  AI_REDIRECT: "redirect",
  AI_X_FACE: "x-face",
  AI_ALTERNATE_NAME: "alternate-name",
  AI_PGP_SIGNATURE: "pgp-signature",
  AI_PGP_PUBLIC_KEY: "pgp-public-key",
  AI_E_MAIL_ADDRESS: "e-mail-address",
  AI_FAQ_TEXT: "faq-text",
  AI_CREATING_SOFTWARE: "creating-software",
  AI_MX_AUTHOR: "mx-author",
  AI_MX_FROM: "mx-from",
  AI_MX_REPLY_TO: "mx-reply-to",
  AI_MX_TO: "mx-to",
  AI_MX_CC: "mx-cc",
  AI_MX_DATE: "mx-date",
  AI_MX_MESSAGE_ID: "mx-message-id",
  AI_MX_IN_REPLY_TO: "mx-in-reply-to",
  AI_MX_MISC: "mx-misc",
  AI_MX_ALLOW_FILTER: "mx-allow-filter",
  AI_MX_REJECT_FORWARD: "mx-reject-forward",
  AI_NOTIFY_COMMENTS: "notify-comments",
  AI_FAQ_FOR_CONF: "faq-for-conf",
  AI_RECOMMENDED_CONF: "recommended-conf",
  AI_ALLOWED_CONTENT_TYPE: "allowed-content-type",
  AI_CANONICAL_NAME: "canonical-name",
  AI_MX_LIST_NAME: "mx-list-name",
  AI_SEND_COMMENTS_TO: "send-comments-to",
  AI_WORLD_READABLE: "world-readable",
  AI_MX_REFUSE_IMPORT: "mx-refuse-import",
  AI_ELISP_CLIENT_READ_FAQ: "elisp-client-read-faq",
  AI_ELISP_CLIENT_REJECTED_RECOMMENDATION: "elisp-client-rejected-recommendation",
  AI_MX_MIME_BELONGS_TO: "mx-mime-belongs-to",
  AI_MX_MIME_PART_IN: "mx-mime-part-in",
  AI_MX_MIME_MISC: "mx-mime-misc",
  AI_MX_ENVELOPE_SENDER: "mx-envelope-sender",
  AI_MX_MIME_FILE_NAME: "mx-mime-file-name",
  AI_MX_ALLOW_ENVELOPE_SENDER_REGEXP: "mx-allow-envelope-sender-regexp",
  AI_KOMFEEDER_UNIQUE_ID: "komfeeder-unique-id",
  AI_KOMFEEDER_URL: "komfeeder-url",
  AI_KOMFEEDER_TITLE: "komfeeder-title",
  AI_KOMFEEDER_AUTHOR: "komfeeder-author",
  AI_KOMFEEDER_DATE: "komfeeder-date",
}



# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.

from __future__ import absolute_import
import errno
import socket
import threading

from .errors import (
    error_dict,
    BadRequestId,
    ReceiveError,
    BadInitialResponse,
    ProtocolError,
    UnimplementedAsync)

from .protocol import (
    to_hstring,
    read_first_non_ws,
    read_int)

from .requests import response_dict
from .asyncmsg import async_dict
from .stats import stats


class ReceiveBuffer(object):
    def __init__(self, socket):
        self._socket = socket

        # Receive buffer
        self._rb = b""    # Buffer for data received from connection
        self._rb_len = 0 # Length of the buffer
        self._rb_pos = 0 # Position of first unread byte in buffer

    def receive_string(self, length):
        """Get a string from the receive buffer (receiving more if
        necessary).
        """
        self._ensure_receive_buffer_size(length)
        res = self._rb[self._rb_pos:self._rb_pos+length]
        self._rb_pos = self._rb_pos + length
        return res

    def receive_char(self):
        """Get a character from the receive buffer (receiving more if
        necessary).
        """
        # FIXME: Optimize for speed
        self._ensure_receive_buffer_size(1)
        res = self._rb[self._rb_pos:self._rb_pos+1]
        self._rb_pos = self._rb_pos + 1
        return res

    def _ensure_receive_buffer_size(self, size):
        """Ensure that there are at least N bytes in the receive
        buffer."""
        # FIXME: Rewrite for speed and clarity
        present = self._rb_len - self._rb_pos
        while present < size:
            needed = size - present
            wanted = max(needed, 128) # FIXME: Optimize
            data = self._socket.recv(wanted)
            if len(data) == 0:
                raise ReceiveError()
            self._rb = self._rb[self._rb_pos:] + data
            self._rb_pos = 0
            self._rb_len = len(self._rb)
            present = self._rb_len


class Connection(object):
    def __init__(self, sock, user=None):
        """

        @param user: See Protocol A spec.
        """
        self._lock = threading.RLock()
        self._socket = sock
        if user is None:
            user = ""
        assert isinstance(user, str) # Do we want user to be str or bytes?

        self._buffer = ReceiveBuffer(self._socket)
        self._ref_no = 0 # Last used ID (i.e. increment before use)
        self._outstanding_requests = {} # Ref-No to Request mapping

        # Send initial string
        self._send_string(b"A%s\n" % (to_hstring(user.encode('latin1')),))

        # Wait for answer "LysKOM\n"
        resp = self._buffer.receive_string(7) # FIXME: receive line here
        if resp != b"LysKOM\n":
            raise BadInitialResponse()
        stats.set('connections.opened.last', 1, agg='sum')

    def send_request(self, req):
        with self._lock:
            ref_no = self._send_request(req)
            stats.set('connections.requests.sent.last', 1, agg='sum')
        return ref_no

    def read_response(self):
        with self._lock:
            ref_no, resp, error = self._parse_response()
        return ref_no, resp, error

    def close(self):
        if self._socket is None:
            return

        with self._lock:
            try:
                self._socket.close()
            except socket.error as e:
                if e.errno in (107, errno.ENOTCONN):
                    # 107: Not connected anymore. Didn't find any errno
                    # name, but the exception says "[Errno 107] Transport
                    # endpoint is not connected".
                    pass
                else:
                    raise
            finally:
                self._socket = None
                stats.set('connections.closed.last', 1, agg='sum')

    def _send_string(self, s):
        """Send a raw string."""
        while len(s) > 0:
            done = self._socket.send(s)
            s = s[done:]

    def _send_request(self, req):
        self._ref_no += 1
        ref_no = self._ref_no
        assert ref_no not in self._outstanding_requests
        request_string = b"%d %s" % (ref_no, req.to_string())
        self._send_string(request_string)
        self._outstanding_requests[ref_no] = req
        return ref_no

    def _parse_response(self):
        ch = read_first_non_ws(self._buffer)
        stats.set('connections.responses.received.last', 1, agg='sum')
        if ch == b"=":
            stats.set('connections.responses.received.ok.last', 1, agg='sum')
            return self._parse_ok_reply()
        elif ch == b"%":
            stats.set('connections.responses.received.error.last', 1, agg='sum')
            return self._parse_error_reply()
        elif ch == b":":
            stats.set('connections.responses.received.async.last', 1, agg='sum')
            return self._parse_asynchronous_message()
        else:
            stats.set('connections.responses.received.protocolerror.last', 1, agg='sum')
            raise ProtocolError("Got unexpected: %s" % (ch,))

    def _parse_ok_reply(self):
        ref_no = read_int(self._buffer)
        if ref_no not in self._outstanding_requests:
            raise BadRequestId(ref_no)
        req = self._outstanding_requests[ref_no]
        resp = response_dict[req.CALL_NO].parse(self._buffer)
        del self._outstanding_requests[ref_no]
        return ref_no, resp, None

    def _parse_error_reply(self):
        ref_no = read_int(self._buffer)
        if ref_no not in self._outstanding_requests:
            raise BadRequestId(ref_no)
        error_no = read_int(self._buffer)
        error_status = read_int(self._buffer)
        error = error_dict[error_no](error_status)
        del self._outstanding_requests[ref_no]
        return ref_no, None, error

    def _parse_asynchronous_message(self):
        read_int(self._buffer) # read number of arguments (but we don't need it)
        msg_no = read_int(self._buffer)
        if msg_no not in async_dict:
            raise UnimplementedAsync(msg_no)
        msg = async_dict[msg_no].parse(self._buffer)
        return None, msg, None



# -*- coding: utf-8 -*-

from __future__ import absolute_import
import six

from . import mimeparse
from .errors import Error
from .protocol import WHITESPACE, DIGITS, ORD_0, to_hstring


def decode_text(text, encoding, backup_encoding='latin1'):
    if encoding is None:
        encoding = backup_encoding
    
    try:
        decoded_text = text.decode(encoding)
    except LookupError:
        # Failed to find the given encoding (such as "x-unkown", for
        # example).
        decoded_text = text.decode(backup_encoding)
    except UnicodeDecodeError:
        # (from iKOM) Fscking clients that can't detect coding...
        decoded_text = text.decode(backup_encoding)
    
    return decoded_text

def parse_content_type(contenttype):
    try:
        mime_type = mimeparse.parse_mime_type(contenttype)
    except Exception:
        raise Error("Failed to parse mime type '%s'" % (contenttype,))
    
    if mime_type[0] == 'x-kom' and mime_type[1] == 'text':
        mime_type = ('text', 'x-kom-basic', mime_type[2])
    
    if "charset" in mime_type[2]:
        # Remove charset from mime_type, if we have it
        encoding = mime_type[2].pop("charset")
    else:
        encoding = None
    
    if encoding == 'x-ctext':
        encoding = 'latin1'
    
    return mime_type, encoding

def mime_type_tuple_to_str(mime_type):
    params = [ "%s=%s" % (k, v) for k, v in mime_type[2].items() ]
    t = "%s/%s" % (mime_type[0], mime_type[1])
    l = [t]
    l.extend(params)
    return ";".join(l)

def parse_hollerith_string(hstring):
    if not hstring:
        return None, hstring
    assert isinstance(hstring, six.binary_type)

    i = 0

    # Skip all leading whitespaces
    c = hstring[i:i+1]
    while c in WHITESPACE:
        i += 1
        c = hstring[i:i+1]

    # Parse length
    length = 0
    while c in DIGITS:
        length = length * 10 + (ord(c) - ORD_0)
        i += 1
        c = hstring[i:i+1]

    if c != b"H":
        raise Exception("Not a valid hollerith string")

    i += 1 # skip "H"
    hend = i + length
    return hstring[i:hend], hstring[hend:]


def to_hollerith_string(s):
    return to_hstring(s)


def decode_user_area(user_area):
    """Decodes the user area to a dictionary where the keys are the
    block names and the values are the hollerith string encoded block
    content.
    """
    h_block_names, h_blocks = parse_hollerith_string(user_area)

    def parse_block_names(hstring):
        # Parse block names
        block_names = []
        h_block_name, hstring = parse_hollerith_string(hstring)
        while h_block_name:
            block_names.append(h_block_name)
            h_block_name, hstring = parse_hollerith_string(hstring)
        return block_names

    def parse_blocks(hstring, block_names):
        blocks = {}
        for block_name in block_names:
            h_block, hstring = parse_hollerith_string(hstring)
            blocks[block_name] = h_block
        return blocks

    block_names = parse_block_names(h_block_names)
    blocks = parse_blocks(h_blocks, block_names)
    return blocks


def encode_user_area(blocks):
    """Encodes a dictionary to a user area. The keys should be the
    block names and the values are the hollerith string encoded
    blocks. The block names will be sorted alphabetically.
    """
    h_block_names = b""
    h_blocks = b""
    for block_name in sorted(blocks.keys()):
        h_block_names += b" " + to_hollerith_string(block_name)
        h_blocks += b" " + to_hollerith_string(blocks[block_name])

    return to_hollerith_string(h_block_names) + h_blocks
